__author__ = 'maartenbreddels'
import PIL.Image
import PIL.ImageDraw
try:
	from StringIO import StringIO
	py3 = False
except ImportError:
	from io import StringIO, BytesIO
	py3 = True
import collections
import numpy as np
import matplotlib.colors
import warnings
from base64 import b64encode

def rgba_2_pil(rgba):
	with warnings.catch_warnings():
		warnings.simplefilter("ignore")
		im = PIL.Image.frombuffer("RGBA", rgba.shape[:2], rgba, 'raw') #, "RGBA", 0, -1)
	return im

def pil_2_data(im, format="png"):
	if py3: # py3 case
		f = BytesIO()
	else:
		f = StringIO()
	im.save(f, format)
	return f.getvalue()

def rgba_to_url(rgba):
	bit8 = rgba.dtype == np.uint8
	if not bit8:
		rgba = (rgba*255.).astype(np.uint8)
	im = rgba_2_pil(rgba)
	data = pil_2_data(im)
	data = b64encode(data)
	if py3:
		data = data.decode("ascii")
	imgurl = "data:image/png;base64," + data + ""
	return imgurl

pdf_modes = collections.OrderedDict()
pdf_modes["multiply"] = np.multiply
pdf_modes["screen"] = lambda a, b: a + b - a * b
pdf_modes["darken"] = np.minimum
pdf_modes["lighten"] = np.maximum

cairo_modes = collections.OrderedDict()
cairo_modes["saturate"] = (lambda aA, aB: np.minimum(1, aA + aB),
						   lambda aA, xA, aB, xB, aR: (np.minimum(aA, 1-aB)*xA + aB*xB)/aR)
cairo_modes["add"] = (lambda aA, aB: np.minimum(1, aA + aB),
						   lambda aA, xA, aB, xB, aR: (aA*xA + aB*xB)/aR)

modes = list(pdf_modes.keys()) + list(cairo_modes.keys())
def background(shape, color="white", alpha=1, bit8=True):
	rgba = np.zeros(shape + (4,))
	rgba[:] = np.array(matplotlib.colors.colorConverter.to_rgba(color))
	rgba[...,3] = alpha
	if bit8:
		return (rgba*255).astype(np.uint8)
	else:
		return rgba


def fade(image_list, opacity=0.5, blend_mode="multiply"):
	result = image_list[0] * 1.
	for i in range(1, len(image_list)):
		result[result[...,3] > 0,3] = 1
		layer = image_list[i] * 1.0
		layer[layer[...,3] > 0,3] = opacity
		result = blend([result, layer], blend_mode=blend_mode)
	return result

def blend(image_list, blend_mode="multiply"):
	bit8 = image_list[0].dtype == np.uint8
	image_list = image_list[::-1]
	rgba_dest = image_list[0] * 1
	if bit8:
		rgba_dest = (rgba_dest/255.).astype(np.float)
	for i in range(1, len(image_list)):
		rgba_source  = image_list[i]
		if bit8:
			rgba_source = (rgba_source/255.).astype(np.float)
		#assert rgba_source.dtype == image_list[0].dtype, "images have different types: first has %r, %d has %r" % (image_list[0].dtype, i, rgba_source.dtype)

		aA = rgba_source[:,:,3]
		aB   = rgba_dest[:,:,3]

		if blend_mode in pdf_modes:
			aR = aA + aB * (1 - aA)
		else:
			aR = cairo_modes[blend_mode][0](aA, aB)
		mask = aR > 0
		for c in range(3): # for r, g and b
			xA = rgba_source[...,c]
			xB = rgba_dest[...,c]
			if blend_mode in pdf_modes:
				f = pdf_modes[blend_mode](xB, xA)
				with np.errstate(divide='ignore', invalid='ignore'): # these are fine, we are ok with nan's in vaex
					result = ((1.-aB) * aA * xA  + (1.-aA) * aB * xB + aA * aB * f) / aR
			else:
				result = cairo_modes[blend_mode][1](aA, xA, aB, xB, aR)
			with np.errstate(divide='ignore', invalid='ignore'): # these are fine, we are ok with nan's in vaex
				result = (np.minimum(aA, 1-aB)*xA + aB*xB)/aR
			#print(result)
			rgba_dest[:,:,c][[mask]] = np.clip(result[[mask]], 0, 1)
		rgba_dest[:,:,3] = np.clip(aR, 0., 1)
	rgba = rgba_dest
	if bit8:
		rgba = (rgba*255).astype(np.uint8)
	return rgba

def monochrome(I, color, vmin=None, vmax=None):
	"""Turns a intensity array to a monochrome 'image' by replacing each intensity by a scaled 'color'

	Values in I between vmin  and vmax get scaled between 0 and 1, and values outside this range are clipped to this.

	Example
	>>> I = np.arange(16.).reshape(4,4)
	>>> color = (0, 0, 1) # red
	>>> rgb = vx.image.monochrome(I, color) # shape is (4,4,3)

	:param I: ndarray of any shape (2d for image)
	:param color: sequence of a (r, g and b) value
	:param vmin: normalization minimum for I, or np.nanmin(I) when None
	:param vmax: normalization maximum for I, or np.nanmax(I) when None
	:return:
	"""
	if vmin is None:
		vmin = np.nanmin(I)
	if vmax is None:
		vmax = np.nanmax(I)
	normalized = (I - vmin) / (vmax - vmin)
	return np.clip(normalized[...,np.newaxis], 0, 1) * np.array(color)

def polychrome(I, colors, vmin=None, vmax=None, axis=-1):
	"""Similar to monochrome, but now do it for multiple colors

	Example
	>>> I = np.arange(32.).reshape(4,4,2)
	>>> colors = [(0, 0, 1), (0, 1, 0)] # red and green
	>>> rgb = vx.image.polychrome(I, colors) # shape is (4,4,3)

	:param I: ndarray of any shape (3d will result in a 2d image)
	:param colors: sequence of [(r,g,b), ...] values
	:param vmin: normalization minimum for I, or np.nanmin(I) when None
	:param vmax: normalization maximum for I, or np.nanmax(I) when None
	:param axis: axis which to sum over, by default the last
	:return:
	"""
	axes_length = len(I.shape)
	allaxes = list(range(axes_length))
	otheraxes = list(allaxes)
	otheraxes.remove((axis + axes_length) % axes_length)
	otheraxes = tuple(otheraxes)

	if vmin is None:
		vmin = np.nanmin(I, axis=otheraxes)
	if vmax is None:
		vmax = np.nanmax(I, axis=otheraxes)
	normalized = (I - vmin) / (vmax - vmin)
	return np.clip(normalized, 0, 1).dot(colors)

#c_r + c_g + c_b



__author__ = 'maartenbreddels'
import os
import mmap
import itertools
import functools
import collections
import logging
import numpy as np
import numpy.ma
import vaex
import astropy.table
import astropy.units
from vaex.utils import ensure_string
import astropy.io.fits as fits
import re
import six

from vaex.dataset import DatasetLocal, DatasetArrays
logger = logging.getLogger("vaex.file")
import vaex.dataset
import vaex.file
dataset_type_map = {}

from vaex.expression import Expression

from vaex.dataset_mmap import DatasetMemoryMapped


import struct
class HansMemoryMapped(DatasetMemoryMapped):
	def __init__(self, filename, filename_extra=None):
		super(HansMemoryMapped, self).__init__(filename)
		self.pageSize, \
		self.formatSize, \
		self.numberParticles, \
		self.numberTimes, \
		self.numberParameters, \
		self.numberCompute, \
		self.dataOffset, \
		self.dataHeaderSize = struct.unpack("Q"*8, self.mapping[:8*8])
		zerooffset = offset = self.dataOffset
		length = self.numberParticles+1
		stride = self.formatSize//8 # stride in units of the size of the element (float64)

		# TODO: ask Hans for the self.numberTimes-2
		lastoffset = offset + (self.numberParticles+1)*(self.numberTimes-2)*self.formatSize
		t_index = 3
		names = "x y z vx vy vz".split()
		midoffset = offset + (self.numberParticles+1)*self.formatSize*t_index
		names = "x y z vx vy vz".split()

		for i, name in enumerate(names):
			self.addColumn(name+"_0", offset+8*i, length, dtype=np.float64, stride=stride)

		for i, name in enumerate(names):
			self.addColumn(name+"_last", lastoffset+8*i, length, dtype=np.float64, stride=stride)


		names = "x y z vx vy vz".split()

		if 1:
			stride = self.formatSize//8
			#stride1 = self.numberTimes #*self.formatSize/8
			for i, name in enumerate(names):
				# TODO: ask Hans for the self.numberTimes-1
				self.addRank1(name, offset+8*i, length=self.numberParticles+1, length1=self.numberTimes-1, dtype=np.float64, stride=stride, stride1=1)

		if filename_extra is None:
			basename = os.path.basename(filename)
			if os.path.exists(basename + ".omega2"):
				filename_extra = basename + ".omega2"

		if filename_extra is not None:
			self.addFile(filename_extra)
			mapping = self.mapping_map[filename_extra]
			names = "J_r J_theta J_phi Theta_r Theta_theta Theta_phi Omega_r Omega_theta Omega_phi r_apo r_peri".split()
			offset = 0
			stride = 11
			#import pdb
			#pdb.set_trace()
			for i, name in enumerate(names):
				# TODO: ask Hans for the self.numberTimes-1
				self.addRank1(name, offset+8*i, length=self.numberParticles+1, length1=self.numberTimes-1, dtype=np.float64, stride=stride, stride1=1, filename=filename_extra)

				self.addColumn(name+"_0", offset+8*i, length, dtype=np.float64, stride=stride, filename=filename_extra)
				self.addColumn(name+"_last", offset+8*i + (self.numberParticles+1)*(self.numberTimes-2)*11*8, length, dtype=np.float64, stride=stride, filename=filename_extra)


	@classmethod
	def can_open(cls, path, *args, **kwargs):
		return os.path.splitext(path)[-1] == ".bin"
		basename, ext = os.path.splitext(path)
		#if os.path.exists(basename + ".omega2"):
		#	return True
		#return True

	@classmethod
	def get_options(cls, path):
		return []

	@classmethod
	def option_to_args(cls, option):
		return []
dataset_type_map["buist"] = HansMemoryMapped

def _python_save_name(name, used=[]):
	first, rest = name[0], name[1:]
	name = re.sub("[^a-zA-Z_]", "_", first) +  re.sub("[^a-zA-Z_0-9]", "_", rest)
	if name in used:
		nr = 1
		while name + ("_%d" % nr) in used:
			nr += 1
		name = name + ("_%d" % nr)
	return name

def _try_unit(unit):
	try:
		unit = astropy.units.Unit(str(unit))
		if not isinstance(unit, astropy.units.UnrecognizedUnit):
			return unit
	except:
		#logger.exception("could not parse unit: %r", unit)
		pass
	try:
		unit_mangle = re.match(".*\[(.*)\]", str(unit)).groups()[0]
		unit = astropy.units.Unit(unit_mangle)
	except:
		pass#logger.exception("could not parse unit: %r", unit)
	if isinstance(unit, six.string_types):
		return None
	elif isinstance(unit, astropy.units.UnrecognizedUnit):
		return None
	else:
		return unit

class FitsBinTable(DatasetMemoryMapped):
	def __init__(self, filename, write=False):
		super(FitsBinTable, self).__init__(filename, write=write)
		with fits.open(filename) as fitsfile:
			for table in fitsfile:
				if isinstance(table, fits.BinTableHDU):
					table_offset = table._data_offset
					#import pdb
					#pdb.set_trace()
					if table.columns[0].dim is not None: # for sure not a colfits
						dim = eval(table.columns[0].dim) # TODO: can we not do an eval here? not so safe
						if len(dim) == 2 and dim[0] <= dim[1]: # we have colfits format
							logger.debug("colfits file!")
							offset = table_offset
							for i in range(len(table.columns)):
								column = table.columns[i]
								cannot_handle = False
								column_name = _python_save_name(column.name.strip(), used=self.columns.keys())
								self._get_column_meta_data(table, column_name, column, i)


								# flatlength == length * arraylength
								flatlength, fitstype = int(column.format[:-1]),column.format[-1]
								arraylength, length = arrayshape = eval(column.dim)

								# numpy dtype code, like f8, i4
								dtypecode = astropy.io.fits.column.FITS2NUMPY[fitstype]


								dtype = np.dtype((">" +dtypecode, arraylength))
								if 0:
									if arraylength > 1:
										dtype = np.dtype((">" +dtypecode, arraylength))
									else:
										if dtypecode == "a": # I think numpy needs by default a length 1
											dtype = np.dtype(dtypecode + "1")
										else:
											dtype = np.dtype(">" +dtypecode)
									#	bytessize = 8

								bytessize = dtype.itemsize
								logger.debug("%r", (column.name, dtype, column.format, column.dim, length, bytessize, arraylength))
								if (flatlength > 0): # and dtypecode != "a": # TODO: support strings
									if dtypecode == "a": # for ascii, we need to add the length again..
										dtypecode += str(arraylength)
									logger.debug("column type: %r", (column.name, offset, dtype, length, column.format, column.dim))
									if arraylength == 1 or dtypecode[0] == "a":
										self.addColumn(column_name, offset=offset, dtype=dtype, length=length)
									else:
										for i in range(arraylength):
											name = column_name+"_" +str(i)
											self.addColumn(name, offset=offset+bytessize*i//arraylength, dtype=">" +dtypecode, length=length, stride=arraylength)
								if flatlength > 0: # flatlength can be
									offset += bytessize * length
								self._check_null(table, column_name, column, i)

					else:
						logger.debug("adding table: %r" % table)
						for i, column in enumerate(table.columns):
							array = column.array[:]
							array = column.array[:] # 2nd time it will be a real np array
							#import pdb
							#pdb.set_trace()
							if array.dtype.kind in "fiubSU":
								column_name = _python_save_name(column.name, used=self.columns.keys())
								self.addColumn(column_name, array=array)
								self._get_column_meta_data(table, column_name, column, i)
								self._check_null(table, column_name, column, i)
			self._try_votable(fitsfile[0])

		self.update_meta()
		self.update_virtual_meta()
		self.selections_favorite_load()

	def _check_null(self, table, column_name, column, i):
		null_name = "TNULL%d" % (i+1)
		if null_name in table.header:
			mask_value = table.header[null_name]
			array = self.columns[column_name]
			mask = array == mask_value
			self.columns[column_name] = numpy.ma.masked_array(array, mask)

	def _try_votable(self, table):
		try:
			from io import BytesIO as StringIO
		except:
			from StringIO import StringIO
		if table.data is None:
			return
		vodata = table.data.tostring()
		if vodata.startswith(b"<?xml"):
			f = StringIO()
			f.write(vodata)
			votable = astropy.io.votable.parse(f)
			first_table = votable.get_first_table()
			used_names = []
			for field in first_table.fields:
				name = field.name.strip()
				clean_name = _python_save_name(name, used=used_names)
				used_names.append(name)
				if field.ucd:
					self.ucds[clean_name] = field.ucd
				unit = _try_unit(field.unit)
				if unit:
					self.units[clean_name] = unit
				if unit is None and field.unit:
					print("unit error for: %r", field.unit)
				self.descriptions[clean_name] = field.description
			self.description = first_table.description

	def _get_column_meta_data(self, table, column_name, column, i):
		ucd_header_name = "TUCD%d" % (i+1)
		if ucd_header_name in table.header:
			self.ucds[column_name] = table.header[ucd_header_name]
		if column.unit:
			try:
				unit = _try_unit(column.unit)
				if unit:
					self.units[column_name] = unit
			except:
				logger.exception("could not understand unit: %s" % column.unit)
		else: # we may want to try ourselves
			unit_header_name = "TUNIT%d" % (i+1)
			if unit_header_name in table.header:
				unit_str = table.header[unit_header_name]
				unit = _try_unit(unit_str)
				if unit:
					self.unit[column_name] = unit
		#unit_header_name = "TUCD%d" % (i+1)
		#if ucd_header_name in table.header:
	@classmethod
	def can_open(cls, path, *args, **kwargs):
		return os.path.splitext(path)[1] == ".fits"

	@classmethod
	def get_options(cls, path):
		return [] # future: support multiple tables?

	@classmethod
	def option_to_args(cls, option):
		return []

dataset_type_map["fits"] = FitsBinTable

# class InMemory(DatasetMemoryMapped):
# 	def __init__(self, name):
# 		super(InMemory, self).__init__(filename=None, nommap=True, name=name)


class SoneiraPeebles(DatasetArrays):
	def __init__(self, dimension, eta, max_level, L):
		super(SoneiraPeebles, self).__init__(name="soneira-peebles")
		#InMemory.__init__(self)
		def todim(value):
			if isinstance(value, (tuple, list)):
				assert len(value) >= dimension, "either a scalar or sequence of length equal to or larger than the dimension"
				return value[:dimension]
			else:
				return [value] * dimension

		eta = eta
		max_level = max_level
		N = eta**(max_level)
		# array[-1] is used as a temp storage
		array = np.zeros((dimension+1, N), dtype=np.float64)
		L = todim(L)

		for d in range(dimension):
			vaex.vaexfast.soneira_peebles(array[d], 0, 1, L[d], eta, max_level)
		for d, name in zip(list(range(dimension)), "x y z w v u".split()):
			self.add_column(name, array[d])
		if 0:
			order = np.zeros(N, dtype=np.int64)
			vaex.vaexfast.shuffled_sequence(order);
			for i, name in zip(list(range(dimension)), "x y z w v u".split()):
				#np.take(array[i], order, out=array[i])
				reorder(array[i], array[-1], order)
				self.addColumn(name, array=array[i])

dataset_type_map["soneira-peebles"] = SoneiraPeebles


class Zeldovich(DatasetArrays):
	def __init__(self, dim=2, N=256, n=-2.5, t=None, seed=None, scale=1, name="zeldovich approximation"):
		super(Zeldovich, self).__init__(name=name)

		if seed is not None:
			np.random.seed(seed)
		#sys.exit(0)
		shape = (N,) * dim
		A = np.random.normal(0.0, 1.0, shape)
		F = np.fft.fftn(A)
		K = np.fft.fftfreq(N, 1./(2*np.pi))[np.indices(shape)]
		k = (K**2).sum(axis=0)
		k_max = np.pi
		F *= np.where(np.sqrt(k) > k_max, 0, np.sqrt(k**n) * np.exp(-k*4.0))
		F.flat[0] = 0
		#pylab.imshow(np.where(sqrt(k) > k_max, 0, np.sqrt(k**-2)), interpolation='nearest')
		grf = np.fft.ifftn(F).real
		Q = np.indices(shape) / float(N-1) - 0.5
		s = np.array(np.gradient(grf)) / float(N)
		#pylab.imshow(s[1], interpolation='nearest')
		#pylab.show()
		s /= s.max() * 100.
		#X = np.zeros((4, 3, N, N, N))
		#for i in range(4):
		#if t is None:
		#	s = s/s.max()
		t = t or 1.
		X = Q + s * t

		for d, name in zip(list(range(dim)), "xyzw"):
			self.add_column(name, X[d].reshape(-1) * scale)
		for d, name in zip(list(range(dim)), "xyzw"):
			self.add_column("v"+name, s[d].reshape(-1) * scale)
		for d, name in zip(list(range(dim)), "xyzw"):
			self.add_column(name+"0", Q[d].reshape(-1) * scale)
		return

dataset_type_map["zeldovich"] = Zeldovich


class AsciiTable(DatasetMemoryMapped):
	def __init__(self, filename):
		super(AsciiTable, self).__init__(filename, nommap=True)
		import asciitable
		table = asciitable.read(filename)
		logger.debug("done parsing ascii table")
		#import pdb
		#pdb.set_trace()
		#names = table.array.dtype.names
		names = table.dtype.names

		#data = table.array.data
		for i in range(len(table.dtype)):
			name = table.dtype.names[i]
			type = table.dtype[i]
			if type.kind in ["f", "i"]: # only store float and int
				#datagroup.create_dataset(name, data=table.array[name].astype(np.float64))
				#dataset.addMemoryColumn(name, table.array[name].astype(np.float64))
				self.addColumn(name, array=table[name])
		#dataset.samp_id = table_id
		#self.list.addDataset(dataset)
		#return dataset

	@classmethod
	def can_open(cls, path, *args, **kwargs):
		can_open = path.endswith(".asc")
		logger.debug("%r can open: %r"  %(cls.__name__, can_open))
		return can_open
dataset_type_map["ascii"] = AsciiTable

class MemoryMappedGadget(DatasetMemoryMapped):
	def __init__(self, filename):
		super(MemoryMappedGadget, self).__init__(filename)
		#h5file = h5py.File(self.filename)
		import vaex.file.gadget
		length, posoffset, veloffset, header = vaex.file.gadget.getinfo(filename)
		self.addColumn("x", posoffset, length, dtype=np.float32, stride=3)
		self.addColumn("y", posoffset+4, length, dtype=np.float32, stride=3)
		self.addColumn("z", posoffset+8, length, dtype=np.float32, stride=3)

		self.addColumn("vx", veloffset, length, dtype=np.float32, stride=3)
		self.addColumn("vy", veloffset+4, length, dtype=np.float32, stride=3)
		self.addColumn("vz", veloffset+8, length, dtype=np.float32, stride=3)
	@classmethod
	def can_open(cls, path, *args, **kwargs):
		with open(path, 'rb') as f:
			first_words = struct.unpack('4I',f.read(4*4))
			if first_words[0] == 8 and first_words[2] == 8 and first_words[3] == 256:
				logg.debug('gadget file SnapFormat=2 detected')
				return True
			elif first_words[0] == 256:
				f.seek(256+4)
				if struct.unpack('I',f.read(4))[0] == 256:
					logger.debug('gadget file SnapFormat=1 detected')
					return True
		return False



dataset_type_map["gadget-plain"] = MemoryMappedGadget

class DatasetAstropyTable(DatasetArrays):
	def __init__(self, filename=None, format=None, table=None, **kwargs):
		if table is None:
			self.filename = filename
			self.format = format
			DatasetArrays.__init__(self, filename)
			self.read_table()
		else:
			#print vars(table)
			#print dir(table)
			DatasetArrays.__init__(self, table.meta.get("name", "unknown-astropy"))
			self.description = table.meta.get("description")
			self.table = table
			#self.name

		#data = table.array.data
		for i in range(len(self.table.dtype)):
			name = self.table.dtype.names[i]
			column = self.table[name]
			type = self.table.dtype[i]
			#clean_name = re.sub("[^a-zA-Z_]", "_", name)
			clean_name = _python_save_name(name, self.columns.keys())
			if type.kind in "fiuSU": # only store float and int
				#datagroup.create_dataset(name, data=table.array[name].astype(np.float64))
				#dataset.addMemoryColumn(name, table.array[name].astype(np.float64))
				masked_array = self.table[name].data
				if "ucd" in column._meta:
					self.ucds[clean_name] = column._meta["ucd"]
				if column.unit:
					unit = _try_unit(column.unit)
					if unit:
						self.units[clean_name] = unit
				if column.description:
					self.descriptions[clean_name] = column.description
				if hasattr(masked_array, "mask"):
					if type.kind in ["f"]:
						masked_array.data[masked_array.mask] = np.nan
					if type.kind in ["i"]:
						masked_array.data[masked_array.mask] = 0
				self.add_column(clean_name, self.table[name].data)
			if type.kind in ["SU"]:
				self.add_column(clean_name, self.table[name].data)

		#dataset.samp_id = table_id
		#self.list.addDataset(dataset)
		#return dataset

	def read_table(self):
		self.table = astropy.table.Table.read(self.filename, format=self.format, **kwargs)

import astropy.io.votable
import string
class VOTable(DatasetArrays):
	def __init__(self, filename):
		DatasetArrays.__init__(self, filename)
		self.filename = filename
		self.path = filename
		votable = astropy.io.votable.parse(self.filename)

		self.first_table = votable.get_first_table()
		self.description = self.first_table.description

		for field in self.first_table.fields:
			name = field.name
			data = self.first_table.array[name]
			type = self.first_table.array[name].dtype
			clean_name = _python_save_name(name, self.columns.keys())
			if field.ucd:
				self.ucds[clean_name] = field.ucd
			if field.unit:
				unit = _try_unit(field.unit)
				if unit:
					self.units[clean_name] = unit
			if field.description:
				self.descriptions[clean_name] = field.description
			if type.kind in "fiubSU": # only store float and int and boolean
				self.add_column(clean_name, data) #self.first_table.array[name].data)
			if type.kind == "O":
				print("column %r is of unsupported object type , will try to convert it to string" % (name,))
				try:
					data = data.astype("S")
					self.add_column(name, data)
				except Exception as e:
					print("Giving up column %s, error: %r" (name, e))
			#if type.kind in ["S"]:
			#	self.add_column(clean_name, self.first_table.array[name].data)

	@classmethod
	def can_open(cls, path, *args, **kwargs):
		can_open = path.endswith(".vot")
		logger.debug("%r can open: %r"  %(cls.__name__, can_open))
		return can_open

dataset_type_map["votable"] = VOTable



class DatasetNed(DatasetAstropyTable):
	def __init__(self, code="2012AJ....144....4M"):
		url = "http://ned.ipac.caltech.edu/cgi-bin/objsearch?refcode={code}&hconst=73&omegam=0.27&omegav=0.73&corr_z=1&out_csys=Equatorial&out_equinox=J2000.0&obj_sort=RA+or+Longitude&of=xml_main&zv_breaker=30000.0&list_limit=5&img_stamp=YES&search_type=Search"\
			.format(code=code)
		super(DatasetNed, self).__init__(url, format="votable", use_names_over_ids=True)
		self.name = "ned:" + code

class DatasetTap(DatasetArrays):
	class TapColumn(object):
		def __init__(self, tap_dataset, column_name, column_type, ucd):
			self.tap_dataset = tap_dataset
			self.column_name = column_name
			self.column_type = column_type
			self.ucd = ucd
			self.alpha_min = 0
			length = len(tap_dataset)
			steps = length/1e6 # try to do it in chunks
			self.alpha_step = 360/steps
			self.alpha_max = self.alpha_min + self.alpha_step
			logger.debug("stepping in alpha %f" % self.alpha_step)
			self.data = []
			self.offset = 0
			self.shape = (length,)
			self.dtype = DatasetTap.type_map[self.column_type]().dtype
			self.left_over_chunk = None
			self.rows_left = length
			import tempfile
			self.download_file = tempfile.mktemp(".vot")

		def __getitem__(self, slice):
			start, stop, step = slice.start, slice.stop, slice.step
			required_length = stop - start
			assert start >= self.offset
			chunk_data = self.left_over_chunk
			enough = False if chunk_data is None else len(chunk_data) >= required_length
			if chunk_data is not None:
				logger.debug("start %s offset %s chunk length %s", start, self.offset, len(chunk_data))
				#assert len(chunk_data) == start - self.offset
			if enough:
				logger.debug("we can skip the query, already have results from previous query")
			while not enough:
				adql_query = "SELECT {column_name} FROM {table_name} WHERE alpha >= {alpha_min} AND alpha < {alpha_max} ORDER BY alpha ASC"\
					.format(column_name=self.column_name, table_name=self.tap_dataset.table_name, alpha_min=self.alpha_min, alpha_max=self.alpha_max)
				logger.debug("executing: %s" % adql_query)
				logger.debug("executing: %s" % adql_query.replace(" ", "+"))


				url = self.tap_dataset.tap_url + "/sync?REQUEST=doQuery&LANG=ADQL&MAXREC=10000000&FORMAT=votable&QUERY=" +adql_query.replace(" ", "+")
				import urllib2
				response = urllib2.urlopen(url)
				with open(self.download_file, "w") as f:
					f.write(response.read())
				votable = astropy.io.votable.parse(self.download_file)
				data = votable.get_first_table().array[self.column_name].data
				# TODO: respect masked array
				#table = astropy.table.Table.read(url, format="votable") #, show_progress=False)
				#data = table[self.column_name].data.data.data
				logger.debug("new chunk is of lenght %d", len(data))
				self.rows_left -= len(data)
				logger.debug("rows left %d", self.rows_left)
				if chunk_data is None:
					chunk_data = data
				else:
					chunk_data = np.concatenate([chunk_data, data])
				if len(chunk_data) >= required_length:
					enough = True
				logger.debug("total chunk is of lenght %d, enough: %s", len(chunk_data), enough)
				self.alpha_min += self.alpha_step
				self.alpha_max += self.alpha_step


			result, self.left_over_chunk = chunk_data[:required_length], chunk_data[required_length:]
			#print(result)
			logger.debug("left over is of length %d", len(self.left_over_chunk))
			return result #np.zeros(N, dtype=self.dtype)



	type_map = {
		'REAL':np.float32,
	    'SMALLINT':np.int32,
		'DOUBLE':np.float64,
		'BIGINT':np.int64,
		'INTEGER':np.int32,
		'BOOLEAN':np.bool8
	}
	#not supported types yet 'VARCHAR',', u'BOOLEAN', u'INTEGER', u'CHAR
	def __init__(self, tap_url="http://gaia.esac.esa.int/tap-server/tap/g10_smc", table_name=None):
		logger.debug("tap url: %r", tap_url)
		self.tap_url = tap_url
		self.table_name = table_name
		if table_name is None: # let us try to infer the table name
			if tap_url.endswith("tap") or tap_url.endswith("tap/"):
				pass # this mean we really didn't provide one
			else:
				index = tap_url.rfind("tap/")
				if index != -1:
					self.tap_url, self.table_name = tap_url[:index+4], self.tap_url[index+4:]
					logger.debug("inferred url is %s, and table name is %s", self.tap_url, self.table_name)

		if self.tap_url.startswith("tap+"): # remove tap+ part from tap+http(s), only keep http(s) part
			self.tap_url = self.tap_url[len("tap+"):]
		import requests
		super(DatasetTap, self).__init__(self.table_name)
		self.req = requests.request("get", self.tap_url+"/tables/")
		self.path = "tap+" +self.tap_url + "/" + table_name

		#print dir(self.req)
		from bs4 import BeautifulSoup
			#self.soup = BeautifulSoup(req.response)
		tables = BeautifulSoup(self.req.content, 'xml')
		self.tap_tables = collections.OrderedDict()
		for table in tables.find_all("table"):
			#print table.find("name").string, table.description.string, table["gaiatap:size"]
			table_name = unicode(table.find("name").string)
			table_size = int(table["esatapplus:size"])
			#print table_name, table_size
			logger.debug("tap table %r ", table_name)
			columns = []
			for column in table.find_all("column"):
				column_name = unicode(column.find("name").string)
				column_type = unicode(column.dataType.string)
				ucd = column.ucd.string if column.ucd else None
				unit = column.unit.string if column.unit else None
				description = column.description.string if column.description else None
				#print "\t", column_name, column_type, ucd
				#types.add()
				columns.append((column_name, column_type, ucd, unit, description))
			self.tap_tables[table_name] = (table_size, columns)
		if not self.tap_tables:
			raise ValueError("no tables or wrong url")
		for name, (table_size, columns) in self.tap_tables.items():
			logger.debug("table %s has length %d", name, table_size)
		self._full_length, self._tap_columns = self.tap_tables[self.table_name]
		self._length = self._full_length
		logger.debug("selected table table %s has length %d", self.table_name, self._full_length)
		#self.column_names = []
		#self.columns = collections.OrderedDict()
		for column_name, column_type, ucd, unit, description in self._tap_columns:
			logger.debug("  column %s has type %s and ucd %s, unit %s and description %s", column_name, column_type, ucd, unit, description)
			if column_type in self.type_map.keys():
				self.column_names.append(column_name)
				if ucd:
					self.ucds[column_name] = ucd
				if unit:
					self.units[column_name] = unit
				if description:
					self.descriptions[column_name] = description
				self.columns[column_name] = self.TapColumn(self, column_name, column_type, ucd)
			else:
				logger.warning("  type of column %s is not supported, it will be skipped", column_name)


	@classmethod
	def can_open(cls, path, *args, **kwargs):
		can_open = False
		url = None
		try:
			url = urlparse(path)
		except:
			return False
		if url.scheme:
			if url.scheme.startswith("tap+http"): # will also catch https
				can_open = True
		logger.debug("%r can open: %r"  %(cls.__name__, can_open))
		return can_open

dataset_type_map["tap"] = DatasetTap



from ._version import __version_tuple__, __version__
versiontuple = __version_tuple__
versionstring = __version__
#if pre_release:
#	versionstring += "-" + pre_release

import platform
# from vaex.utils import osname, setup.py doesn't want imports...
osname = dict(darwin="osx", linux="linux", windows="windows")[platform.system().lower()]

if __name__ == "__main__":
	import vaex
	import sys
	#print vaex.__version_tuple__
	if sys.argv[1] == "version":
		print("version:", vaex.__version__)
	elif sys.argv[1] == "fullname":
		print("full name:", vaex.__full_name__)
	elif sys.argv[1] == "buildname":
		print("build name:", vaex.__build_name__)
	elif sys.argv[1] == "tagcmd":
		print("git tag %s" % versionstring)
		print("git push --tags")
	else:
		print("use version, fullname or buildname as argument")



# -*- coding: utf-8 -*-
from __future__ import division, print_function
import numpy as np
import vaex
from .dataset import Task, TaskMapReduce, _parse_f
import six

def _asfloat(a):
	if a.dtype.type==np.float64 and a.strides[0] == 8:
		return a
	else:
		return a.astype(np.float64, copy=False)


class TaskHistogram(Task):
	def __init__(self, dataset, subspace, expressions, size, limits, masked=False, weight=None):
		self.size = size
		self.limits = limits
		Task.__init__(self, dataset, expressions, name="histogram")
		self.subspace = subspace
		self.dtype = np.float64
		self.masked = masked
		self.weight = weight
		#self.grids = vaex.grids.Grids(self.dataset, self.dataset.executor.thread_pool, *expressions)
		#self.grids.ranges = limits
		#self.grids.grids["counts"] = vaex.grids.Grid(self.grids, size, self.dimension, None)
		shape1 = ( self.size,) * self.dimension
		try:
			self.size[0]
			shape1 = tuple(self.size)
		except:
			pass
		shape = (self.subspace.executor.thread_pool.nthreads,) + shape1
		self.data = np.zeros(shape, dtype=self.dtype)
		self.ranges_flat = []
		self.minima = []
		self.maxima = []
		for limit in self.limits:
			self.ranges_flat.extend(limit)
			vmin, vmax = limit
			self.minima.append(vmin)
			self.maxima.append(vmax)
		if self.weight is not None:
			self.expressions_all.append(weight)
		#print self.ranges_flat

	def __repr__(self):
		name = self.__class__.__module__ + "." +self.__class__.__name__
		return "<%s(dataset=%r, expressions=%r, size=%r, limits=%r)> instance at 0x%x" % (name, self.dataset, self.expressions, self.size, self.limits, id(self))

	def map(self, thread_index, i1, i2, *blocks):
		class Info(object):
			pass
		info = Info()
		info.i1 = i1
		info.i2 = i2
		info.first = i1 == 0
		info.last = i2 == self.dataset.length_unfiltered()
		info.size = i2-i1
		#print "bin", i1, i2, info.last
		#self.grids["counts"].bin_block(info, *blocks)
		#mask = self.dataset.mask
		data = self.data[thread_index]

		blocks = [_asfloat(block) for block in blocks]

		if self.masked or self.dataset.filtered:
			mask = self.dataset.evaluate_selection_mask("default" if self.masked else None, i1=i1, i2=i2)
			blocks = [block[mask] for block in blocks]

		subblock_weight = None
		if len(blocks) == len(self.expressions) + 1:
			subblock_weight = blocks[-1]
			blocks = list(blocks[:-1])
		#print subblocks[0]
		#print subblocks[1]

		if self.dimension == 1:
			vaex.vaexfast.histogram1d(blocks[0], subblock_weight, data, *self.ranges_flat)
		elif self.dimension == 2:
			#if subblock_weight is None:
			#	#print "speedup?"
			#	histogram_numba(blocks[0], blocks[1], subblock_weight, data, *self.ranges_flat)
			#else:
			vaex.vaexfast.histogram2d(blocks[0], blocks[1], subblock_weight, data, *self.ranges_flat)
			#vaex.vaexfast.statisticNd([blocks[0], blocks[1]], subblock_weight, data, self.minima, self.maxima, 0)
		elif self.dimension == 3:
			vaex.vaexfast.histogram3d(blocks[0], blocks[1], blocks[2], subblock_weight, data, *self.ranges_flat)
		else:
			blocks = list(blocks) # histogramNd wants blocks to be a list
			vaex.vaexfast.histogramNd(blocks, subblock_weight, data, self.minima, self.maxima)

		return i1
		#return map(self._map, blocks)#[self.map(block) for block in blocks]

	def reduce(self, results):
		for i in range(1, self.subspace.executor.thread_pool.nthreads):
			self.data[0] += self.data[i]
		return self.data[0]
		#return self.data


class SubspaceGridded(object):
	def __init__(self, subspace_bounded, grid, vx=None, vy=None, vcounts=None):
		self.subspace_bounded = subspace_bounded
		self.grid = grid
		self.vx = vx
		self.vy = vy
		self.vcounts = vcounts

	def vector(self, weightx, weighty, size=32):
		counts = self.subspace_bounded.gridded_by_histogram(size=size)
		vx = self.subspace_bounded.gridded_by_histogram(size=size, weight=weightx)
		vy = self.subspace_bounded.gridded_by_histogram(size=size, weight=weighty)
		return SubspaceGridded(self.subspace_bounded, self.grid, vx=vx, vy=vy, vcounts=counts)

	def filter_gaussian(self, sigmas=1):
		import scipy.ndimage
		return SubspaceGridded(self.subspace_bounded, scipy.ndimage.filters.gaussian_filter(self.grid, sigmas))

	def clip_relative(self, v1, v2):
		vmin = self.grid.min()
		vmax = self.grid.max()
		width = vmax - vmin
		return SubspaceGridded(self.subspace_bounded, np.clip(self.grid, vmin + v1 * width, vmin + v2 * width))

	def volr(self,  **kwargs):
		import vaex.notebook
		return vaex.notebook.volr(subspace_gridded=self, **kwargs)

	def plot(self, axes=None, **kwargs):
		self.subspace_bounded.subspace.plot(np.log1p(self.grid), limits=self.subspace_bounded.bounds, axes=axes, **kwargs)

	def mean_line(self, axis=0, **kwargs):
		from matplotlib import pylab
		assert axis in [0,1]
		other_axis = 0 if axis == 1 else 1
		xmin, xmax = self.subspace_bounded.bounds[axis]
		ymin, ymax = self.subspace_bounded.bounds[other_axis]
		x = vaex.utils.linspace_centers(xmin, xmax, self.grid.shape[axis])
		y = vaex.utils.linspace_centers(ymin, ymax, self.grid.shape[other_axis])
		print(y)
		if axis == 0:
			counts = np.sum(self.grid, axis=axis)
			means = np.sum(self.grid * y[np.newaxis,:].T, axis=axis)/counts
		else:
			counts = np.sum(self.grid, axis=axis)
			means = np.sum(self.grid * y[:,np.newaxis].T, axis=axis)/counts
		if axis == 0:
			result = pylab.plot(x, means, **kwargs)
		else:
			result = pylab.plot(means, x, **kwargs)

		self.subspace_bounded.lim()
		return result, x, means



	def _repr_png_(self):
		from matplotlib import pylab
		fig, ax = pylab.subplots()
		self.plot(axes=ax, f=np.log1p)
		import vaex.utils
		if all([k is not None for k in [self.vx, self.vy, self.vcounts]]):
			N = self.vx.grid.shape[0]
			bounds = self.subspace_bounded.bounds
			print(bounds)
			positions = [vaex.utils.linspace_centers(bounds[i][0], bounds[i][1], N) for i in range(self.subspace_bounded.subspace.dimension)]
			print(positions)
			mask = self.vcounts.grid > 0
			vx = np.zeros_like(self.vx.grid)
			vy = np.zeros_like(self.vy.grid)
			vx[mask] = self.vx.grid[mask] / self.vcounts.grid[mask]
			vy[mask] = self.vy.grid[mask] / self.vcounts.grid[mask]
			#vx = self.vx.grid / self.vcounts.grid
			#vy = self.vy.grid / self.vcounts.grid
			x2d, y2d = np.meshgrid(positions[0], positions[1])
			ax.quiver(x2d[mask], y2d[mask], vx[mask], vy[mask])
			#print x2d
			#print y2d
			#print vx
			#print vy
			#ax.quiver(x2d, y2d, vx, vy)
		ax.title.set_text("$\log(1+counts)$")
		ax.set_xlabel(self.subspace_bounded.subspace.expressions[0])
		ax.set_ylabel(self.subspace_bounded.subspace.expressions[1])
		#pylab.savefig
		#from .io import StringIO
		from six import StringIO
		file_object = StringIO()
		fig.canvas.print_png(file_object)
		pylab.close(fig)
		return file_object.getvalue()

	def cube_png(self, f=np.log1p, colormap="afmhot", file="cube.png"):
		if self.grid.shape != ((128,) *3):
			logger.error("only 128**3 cubes are supported")
			return None
		colormap_name = "afmhot"
		import matplotlib.cm
		colormap = matplotlib.cm.get_cmap(colormap_name)
		mapping = matplotlib.cm.ScalarMappable(cmap=colormap)
		#pixmap = QtGui.QPixmap(32*2, 32)
		data = np.zeros((128*8, 128*16, 4), dtype=np.uint8)

		#mi, ma = 1*10**self.mod1, self.data3d.max()*10**self.mod2
		grid = f(self.grid)
		vmin, vmax = grid.min(), grid.max()
		grid_normalized = (grid-vmin) / (vmax-vmin)
		#intensity_normalized = (np.log(self.data3d + 1.) - np.log(mi)) / (np.log(ma) - np.log(mi));
		import PIL.Image
		for y2d in range(8):
			for x2d in range(16):
				zindex = x2d + y2d*16
				I = grid_normalized[zindex]
				rgba = mapping.to_rgba(I,bytes=True) #.reshape(Nx, 4)
				#print rgba.shape
				subdata = data[y2d*128:(y2d+1)*128, x2d*128:(x2d+1)*128]
				for i in range(3):
					subdata[:,:,i] = rgba[:,:,i]
				subdata[:,:,3] = (grid_normalized[zindex]*255).astype(np.uint8)# * 0 + 255
				if 0:
					filename = "cube%03d.png" % zindex
					img = PIL.Image.frombuffer("RGB", (128, 128), subdata[:,:,0:3] * 1)
					print(("saving to", filename))
					img.save(filename)
		img = PIL.Image.frombuffer("RGBA", (128*16, 128*8), data, 'raw') #, "RGBA", 0, -1)
		#filename = "cube.png"
		#print "saving to", file
		img.save(file, "png")

		if 0:
			filename = "colormap.png"
			print(("saving to", filename))
			height, width = self.colormap_data.shape[:2]
			img = PIL.Image.frombuffer("RGB", (width, height), self.colormap_data)
			img.save(filename)


class SubspaceBounded(object):
	def __init__(self, subspace, bounds):
		self.subspace = subspace
		self.bounds = bounds

	def histogram(self, size=256, weight=None):
		return self.subspace.histogram(limits=self.bounds, size=size, weight=weight)

	def gridded(self, size=256, weight=None):
		return self.gridded_by_histogram(size=size, weight=weight)

	def gridded_by_histogram(self, size=256, weight=None):
		grid = self.histogram(size=size, weight=weight)
		return SubspaceGridded(self, grid)

	def lim(self):
		from matplotlib import pylab
		xmin, xmax = self.bounds[0]
		ymin, ymax = self.bounds[1]
		pylab.xlim(xmin, xmax)
		pylab.ylim(ymin, ymax)


class Subspaces(object):
	"""
	:type: subspaces: list[Subspace]

	"""
	def __init__(self, subspaces):
		self.subspaces = subspaces
		self.expressions = set()
		first_subspace = self.subspaces[0]
		self.delay = first_subspace.delay
		self.dimension = first_subspace.dimension
		self.dataset = self.subspaces[0].dataset
		for subspace in self.subspaces:
			assert subspace.dataset == self.subspaces[0].dataset
			assert subspace.delay == self.subspaces[0].delay
			assert subspace.dimension == self.subspaces[0].dimension, "subspace is of dimension %s, while first subspace if of dimension %s" % (subspace.dimension, self.subspaces[0].dimension)
			#assert subspace.sele== self.subspaces[0].delay
			self.expressions.update(subspace.expressions)
		self.expressions = list(self.expressions)
		self.subspace = self.dataset(*list(self.expressions), delay=self.delay, executor=first_subspace.executor)

	#def _repr_html_(self):



	def __len__(self):
		return len(self.subspaces)

	def names(self, seperator=" "):
		return [seperator.join(subspace.expressions) for subspace in self.subspaces]

	def expressions_list(self):
		return [subspace.expressions for subspace in self.subspaces]

	def selected(self):
		return Subspaces([subspace.selected() for subspace in self.subspaces])

	def _unpack(self, values):
		value_map = dict(zip(self.expressions, values))
		return [[value_map[ex] for ex in subspace.expressions] for subspace in self.subspaces]

	def _pack(self, values):
		value_map = {}
		for subspace_values, subspace in zip(values, self.subspaces):
			for value, expression in zip(subspace_values, subspace.expressions):
				if expression in value_map:
					if isinstance(value, np.ndarray):
						assert np.all(value_map[expression] == value), "inconsistency in subspaces, value for expression %r is %r in one case, and %r in the other" % (expression, value, value_map[expression])
					else:
						assert value_map[expression] == value, "inconsistency in subspaces, value for expression %r is %r in one case, and %r in the other" % (expression, value, value_map[expression])
				else:
					value_map[expression] = value
		return [value_map[expression] for expression in self.expressions]


	def minmax(self):
		if self.delay:
			return self.subspace.minmax().then(self._unpack)
		else:
			return self._unpack(self.subspace.minmax())

	def limits_sigma(self, sigmas=3, square=False):
		if self.delay:
			return self.subspace.limits_sigma(sigmas=sigmas, square=square).then(self._unpack)
		else:
			return self._unpack(self.subspace.limits_sigma(sigmas=sigmas, square=square))

	def mutual_information(self, limits=None, size=256):
		if limits is not None:
			limits = self._pack(limits)
		def mutual_information(limits):
			return vaex.promise.listPromise([vaex.promise.Promise.fulfilled(subspace.mutual_information(subspace_limits, size=size)) for subspace_limits, subspace in zip(limits, self.subspaces)])
			#return histograms
		if limits is None:
			limits_promise = vaex.promise.Promise.fulfilled(self.subspace.minmax())
		else:
			limits_promise = vaex.promise.Promise.fulfilled(limits)
		limits_promise = limits_promise.then(self._unpack)
		promise = limits_promise.then(mutual_information)
		return promise if self.delay else promise.get()



	def mean(self):
		if self.delay:
			return self.subspace.mean().then(self._unpack)
		else:
			means = self.subspace.mean()
			return self._unpack(means)

	def var(self, means=None):
		# 'pack' means, and check if it makes sence
		if means is not None:
			means = self._pack(means)
		def var(means):
			return self.subspace.var(means=means)
		if self.delay:
			#if means is None:
			#	return self.subspace.mean().then(var).then(self._unpack)
			#else:
			return var(means).then(self._unpack)
		else:
			#if means is None:
			#	means = self.subspace.mean()
			#logger.debug("means: %r", means)
			return self._unpack(var(means=means))

	def correlation(self, means=None, vars=None):
		def var(means):
			return self.subspace.var(means=means)
		def correlation(means_and_vars):
			means, vars = means_and_vars
			means, vars = self._unpack(means), self._unpack(vars)
			#return self.subspace.correlation(means=means, vars=vars)
			return vaex.promise.listPromise([subspace.correlation(means=subspace_mean, vars=subspace_var) for subspace_mean, subspace_var, subspace in zip(means, vars, self.subspaces)])
		if means is not None:
			means = self._pack(means)
		if vars is not None:
			vars = self._pack(vars)
		if self.delay:
			if means is None:
				mean_promise = self.subspace.mean()
			else:
				mean_promise = vaex.promise.Promise.fulfilled(means)
			if vars is None:
				var_promise = mean_promise.then(var)
			else:
				var_promise = vaex.promise.Promise.fulfilled(vars)
			mean_and_var_calculated = vaex.promise.listPromise(mean_promise, var_promise)
			return mean_and_var_calculated.then(correlation)
		else:
			if means is None:
				means = self.subspace.mean()
			if vars is None:
				vars = self.subspace.var(means=means)
			means = self._unpack(means)
			vars = self._unpack(vars)
			return [subspace.correlation(means=subspace_mean, vars=subspace_var) for subspace_mean, subspace_var, subspace in zip(means, vars, self.subspaces)]
			#return correlation((means, vars))





	#def bounded_by(self, limits_list):
	#	return SubspacesBounded(SubspaceBounded(subspace, limits) for subspace, limit in zip(self.subspaces, limits_list))

class Subspace(object):
	"""A Subspace represent a subset of columns or expressions from a dataset.

	subspace are not instantiated directly, but by 'calling' the dataset like this:

	>>> subspace_xy = some_dataset("x", "y")
	>>> subspace_r = some_dataset("sqrt(x**2+y**2)")

	See `vaex.dataset.Dataset` for more documentation.

	"""
	def __init__(self, dataset, expressions, executor, delay, masked=False):
		"""

		:param Dataset dataset: the dataset the subspace refers to
		:param list[str] expressions: list of expressions that forms the subspace
		:param Executor executor: responsible for executing the tasks
		:param bool delay: return answers directly, or as a promise
		:param bool masked: work on the selection or not
		:return:
		"""
		self.dataset = dataset
		self.expressions = expressions
		self.executor = executor
		self.delay = delay
		self.is_masked = masked

	def __repr__(self):
		name = self.__class__.__module__ + "." +self.__class__.__name__
		return "<%s(dataset=%r, expressions=%r, delay=%r, is_masked=%r)> instance at 0x%x" % (name, self.dataset, self.expressions, self.delay, self.is_masked, id(self))

	@property
	def dimension(self):
		return len(self.expressions)

	def get_selection(self):
		return self.dataset.get_selection("default") if self.is_masked else None

	def is_selected(self):
		return self.is_masked

	def selected(self):
		return self.__class__(self.dataset, expressions=self.expressions, executor=self.executor, delay=self.delay, masked=True)

	def delayhronous(self):
		return self.__class__(self.dataset, expressions=self.expressions, executor=self.executor, delay=True, masked=self.is_masked)

	def image_rgba_save(self, filename, data=None, rgba8=None, **kwargs):
		if rgba8 is not None:
			data = self.image_rgba_data(rgba8=rgba8, **kwargs)
		if data is None:
			data = self.image_rgba_data(**kwargs)
		with open(filename, "wb") as f:
			f.write(data)

	def image_rgba_notebook(self, data=None, rgba8=None, **kwargs):
		if rgba8 is not None:
			data = self.image_rgba_data(rgba8=rgba8, **kwargs)
		if data is None:
			data = self.image_rgba_data(**kwargs)
		from IPython.display import display, Image
		return Image(data=data)

	def image_rgba_data(self, rgba8=None, format="png", pil_draw=False, **kwargs):
		import PIL.Image
		import PIL.ImageDraw
		from six import StringIO
		if rgba8 is None:
			rgba8 = self.image_rgba(**kwargs)
		img = PIL.Image.frombuffer("RGBA", rgba8.shape[:2], rgba8, 'raw') #, "RGBA", 0, -1)
		if pil_draw:
			draw = PIL.ImageDraw.Draw(img)
			pil_draw(draw)

		f = StringIO()
		img.save(f, format)
		return f.getvalue()

	def image_rgba_url(self, rgba8=None, **kwargs):
		if rgba8 is None:
			rgba8 = self.image_rgba(**kwargs)
		import PIL.Image
		img = PIL.Image.frombuffer("RGBA", rgba8.shape[:2], rgba8, 'raw') #, "RGBA", 0, -1)
		from six import StringIO
		f = StringIO()
		img.save(f, "png")
		from base64 import b64encode
		imgurl = "data:image/png;base64," + b64encode(f.getvalue()) + ""
		return imgurl

	def normalize_grid(self, grid):
		grid = grid * 1 # copy
		mask = (grid > 0) & np.isfinite(grid)
		if grid.sum():
			grid -= grid[mask].min()
			grid /= grid[mask].max()
		else:
			grid[:] = 0
		return grid

	def limits(self, value, square=False):
		"""TODO: doc + server side implementation"""
		if isinstance(value, six.string_types):
			import re
			match = re.match("(\d*)(\D*)", value)
			if match is None:
				raise ValueError("do not understand limit specifier %r, examples are 90%, 3sigma")
			else:
				value, type = match.groups()
				import ast
				value = ast.literal_eval(value)
				type = type.strip()
				if type in ["s", "sigma"]:
					return self.limits_sigma(value)
				elif type in ["ss", "sigmasquare"]:
					return self.limits_sigma(value, square=True)
				elif type in ["%", "percent"]:
					return self.limits_percentage(value)
				elif type in ["%s", "%square", "percentsquare"]:
					return self.limits_percentage(value, square=True)
		if value is None:
			return self.limits_percentage(square=square)
		else:
			return value

	def image_rgba(self, grid=None, size=256, limits=None, square=False, center=None, weight=None, weight_stat="mean", figsize=None,
			 aspect="auto", f=lambda x: x, axes=None, xlabel=None, ylabel=None,
			 group_by=None, group_limits=None, group_colors='jet', group_labels=None, group_count=10, cmap="afmhot",
		     vmin=None, vmax=None,
			 pre_blend=False, background_color="white", background_alpha=1., normalize=True, color=None):
		f = _parse_f(f)
		if grid is None:
			limits = self.limits(limits)
			if limits is None:
				limits = self.limits_sigma()
			if group_limits is None and group_by:
				group_limits = tuple(self.dataset(group_by).minmax()[0]) + (group_count,)
			if weight_stat == "mean" and weight is not None:
				grid = self.bin_mean(weight, limits=limits, size=size, group_limits=group_limits, group_by=group_by)
			else:
				grid = self.histogram(limits=limits, size=size, weight=weight, group_limits=group_limits, group_by=group_by)
			if grid is None: # cancel occured
				return
		import matplotlib.cm
		background_color = np.array(matplotlib.colors.colorConverter.to_rgb(background_color))
		if group_by:
			gmin, gmax, group_count = group_limits
			if isinstance(group_colors, six.string_types):
				group_colors = matplotlib.cm.get_cmap(group_colors)
			if isinstance(group_colors, matplotlib.colors.Colormap):
				group_count = group_limits[2]
				colors = [group_colors(k/float(group_count-1.)) for k in range(group_count) ]
			else:
				colors = [matplotlib.colors.colorConverter.to_rgba(k) for k in group_colors]
			total = np.sum(grid, axis=0).T
			#grid /= total
			mask = total > 0
			alpha = total - total[mask].min()
			alpha[~mask] = 0
			alpha = total / alpha.max()
			rgba = grid.T.dot(colors)
			def _norm(data):
				mask = np.isfinite(data)
				data = data - data[mask].min()
				data /= data[mask].max()
				return data
			rgba[...,3] = (f(alpha))
			#rgba[...,3] = 1
			rgba[total == 0,3] = 0.
			mask = alpha > 0
			if 1:
				for i in range(3):
					rgba[...,i] /= total
					#rgba[...,i] /= rgba[...,0:3].max()
					rgba[~mask,i] = background_color[i]
			rgba = (np.swapaxes(rgba, 0, 1))
		else:
			if color:
				color = np.array(matplotlib.colors.colorConverter.to_rgba(color))
				rgba = np.zeros(grid.shape + (4,))
				rgba[...,0:4] = color
				data = f(grid)
				mask = (grid > 0) & np.isfinite(data)
				if vmin is None:
					vmin = data[mask].min()
				if vmax is None:
					vmax = data[mask].max()
				if mask.sum():
					data -= vmin
					data /= vmax
					data[~mask] = 0
				else:
					data[:] = 0
				rgba[...,3] = data
			else:
				cmap = matplotlib.cm.get_cmap(cmap)
				data = f(grid)
				if normalize:
					mask = (data > 0) & np.isfinite(data)
					if vmin is None:
						vmin = data[mask].min()
					if vmax is None:
						vmax = data[mask].max()
					if mask.sum():
						data -= vmin
						data /= vmax
					else:
						data[:] = 0
					data[~mask] = 0
				data = np.clip(data, 0, 1)
				rgba = cmap(data)
				if normalize:
					rgba[~mask,3] = 0
				rgba[...,3] = 1#data
			#rgba8 = np.swapaxes(rgba8, 0, 1)
		#white = np.ones_like(rgba[...,0:3])
		if pre_blend:
			#rgba[...,3] = background_alpha
			rgb = rgba[...,:3].T
			alpha = rgba[...,3].T
			rgb[:] = rgb * alpha + background_color[:3].reshape(3,1,1) * (1-alpha)
			alpha[:] = alpha + background_alpha * (1-alpha)
		rgba= np.clip(rgba, 0, 1)
		rgba8 = (rgba*255).astype(np.uint8)
		return rgba8

	def plot_vectors(self, expression_x, expression_y, limits, wx=None, wy=None, counts=None, size=32, axes=None, **kwargs):
		import pylab
		# refactor: should go to bin_means_xy
		if counts is None:
			counts = self.histogram(size=size, limits=limits)
		if wx is None:
			wx = self.histogram(size=size, weight=expression_x, limits=limits)
		if wy is None:
			wy = self.histogram(size=size, weight=expression_y, limits=limits)
		N = size
		positions = [vaex.utils.linspace_centers(limits[i][0], limits[i][1], N) for i in range(self.dimension)]
		#print(positions)
		mask = counts > 0
		vx = wx/counts
		vy = wy/counts
		vx[counts==0] = 0
		vy[counts==0] = 0
		#vx = self.vx.grid / self.vcounts.grid
		#vy = self.vy.grid / self.vcounts.grid
		x2d, y2d = np.meshgrid(positions[0], positions[1])
		if axes is None:
			axes = pylab.gca()
		axes.quiver(x2d[mask], y2d[mask], vx[mask], vy[mask], **kwargs)

	def plot(self, grid=None, size=256, limits=None, square=False, center=None, weight=None, weight_stat="mean", figsize=None,
			 aspect="auto", f="identity", axes=None, xlabel=None, ylabel=None,
			 group_by=None, group_limits=None, group_colors='jet', group_labels=None, group_count=None,
			 vmin=None, vmax=None,
			 cmap="afmhot",
			 **kwargs):
		"""Plot the subspace using sane defaults to get a quick look at the data.

		:param grid: A 2d numpy array with the counts, if None it will be calculated using limits provided and Subspace.histogram
		:param size: Passed to Subspace.histogram
		:param limits: Limits for the subspace in the form [[xmin, xmax], [ymin, ymax]], if None it will be calculated using Subspace.limits_sigma
		:param square: argument passed to Subspace.limits_sigma
		:param Executor executor: responsible for executing the tasks
		:param figsize: (x, y) tuple passed to pylab.figure for setting the figure size
		:param aspect: Passed to matplotlib's axes.set_aspect
		:param xlabel: String for label on x axis (may contain latex)
		:param ylabel: Same for y axis
		:param kwargs: extra argument passed to axes.imshow, useful for setting the colormap for instance, e.g. cmap='afmhot'
		:return: matplotlib.image.AxesImage

		 """
		import pylab
		f = _parse_f(f)
		limits = self.limits(limits)
		if limits is None:
			limits = self.limits_sigma()
		#if grid is None:
		if group_limits is None and group_by:
			group_limits = tuple(self.dataset(group_by).minmax()[0]) + (group_count,)
		#	grid = self.histogram(limits=limits, size=size, weight=weight, group_limits=group_limits, group_by=group_by)
		if figsize is not None:
			pylab.figure(num=None, figsize=figsize, dpi=80, facecolor='w', edgecolor='k')
		if axes is None:
			axes = pylab.gca()
		fig = pylab.gcf()
		#if xlabel:
		pylab.xlabel(xlabel or self.expressions[0])
		#if ylabel:
		pylab.ylabel(ylabel or self.expressions[1])
		#axes.set_aspect(aspect)
		rgba8 = self.image_rgba(grid=grid, size=size, limits=limits, square=square, center=center, weight=weight, weight_stat=weight_stat,
			 f=f, axes=axes,
			 group_by=group_by, group_limits=group_limits, group_colors=group_colors, group_count=group_count,
			vmin=vmin, vmax=vmax,
			 cmap=cmap)
		import matplotlib
		if group_by:
			if isinstance(group_colors, six.string_types):
				group_colors = matplotlib.cm.get_cmap(group_colors)
			if isinstance(group_colors, matplotlib.colors.Colormap):
				group_count = group_limits[2]
				colors = [group_colors(k/float(group_count-1.)) for k in range(group_count) ]
			else:
				colors = [matplotlib.colors.colorConverter.to_rgba(k) for k in group_colors]
			colormap = matplotlib.colors.ListedColormap(colors)
			gmin, gmax, group_count = group_limits#[:2]
			delta = (gmax - gmin) / (group_count-1.)
			norm = matplotlib.colors.Normalize(gmin-delta/2, gmax+delta/2)
			sm = matplotlib.cm.ScalarMappable(norm, colormap)
			sm.set_array(1) # make matplotlib happy (strange behavious)
			colorbar = fig.colorbar(sm)
			if group_labels:
				colorbar.set_ticks(np.arange(gmin, gmax+delta/2, delta))
				colorbar.set_ticklabels(group_labels)
			else:
				colorbar.set_ticks(np.arange(gmin, gmax+delta/2, delta))
				colorbar.set_ticklabels(map(lambda x: "%f" % x, np.arange(gmin, gmax+delta/2, delta)))
			colorbar.ax.set_ylabel(group_by)
			#matplotlib.colorbar.ColorbarBase(axes, norm=norm, cmap=colormap)
			im = axes.imshow(rgba8, extent=np.array(limits).flatten(), origin="lower", aspect=aspect, **kwargs)
		else:
			norm = matplotlib.colors.Normalize(0, 23)
			sm = matplotlib.cm.ScalarMappable(norm, cmap)
			sm.set_array(1) # make matplotlib happy (strange behavious)
			colorbar = fig.colorbar(sm)
			im = axes.imshow(rgba8, extent=np.array(limits).flatten(), origin="lower", aspect=aspect, **kwargs)
			colorbar = None
		return im, colorbar

	def plot1d(self, grid=None, size=64, limits=None, weight=None, figsize=None, f="identity", axes=None, xlabel=None, ylabel=None, **kwargs):
		"""Plot the subspace using sane defaults to get a quick look at the data.

		:param grid: A 2d numpy array with the counts, if None it will be calculated using limits provided and Subspace.histogram
		:param size: Passed to Subspace.histogram
		:param limits: Limits for the subspace in the form [[xmin, xmax], [ymin, ymax]], if None it will be calculated using Subspace.limits_sigma
		:param figsize: (x, y) tuple passed to pylab.figure for setting the figure size
		:param xlabel: String for label on x axis (may contain latex)
		:param ylabel: Same for y axis
		:param kwargs: extra argument passed to ...,

		 """
		import pylab
		f = _parse_f(f)
		limits = self.limits(limits)
		assert self.dimension == 1, "can only plot 1d, not %s" % self.dimension
		if limits is None:
			limits = self.limits_sigma()
		if grid is None:
			grid = self.histogram(limits=limits, size=size, weight=weight)
		if figsize is not None:
			pylab.figure(num=None, figsize=figsize, dpi=80, facecolor='w', edgecolor='k')
		if axes is None:
			axes = pylab.gca()
		#if xlabel:
		pylab.xlabel(xlabel or self.expressions[0])
		#if ylabel:
		#pylab.ylabel(ylabel or self.expressions[1])
		pylab.ylabel("counts" or ylabel)
		#axes.set_aspect(aspect)
		N = len(grid)
		xmin, xmax = limits[0]
		return pylab.plot(np.arange(N) / (N-1.0) * (xmax-xmin) + xmin, f(grid,), drawstyle="steps", **kwargs)
		#pylab.ylim(-1, 6)


	def plot_histogram_bq(self, f="identity", size=64, limits=None, color="red", bq_cleanup=True):
		import vaex.ext.bqplot
		limits = self.limits(limits)
		plot = vaex.ext.bqplot.BqplotHistogram(self, color, size, limits)
		if not hasattr(self, "_bqplot"):
			self._bqplot = {}
			self._bqplot["cleanups"] = []
		else:
			if bq_cleanup:
				for cleanup in self._bqplot["cleanups"]:
					cleanup()
			self._bqplot["cleanups"] = []

		def cleanup(callback=plot.callback):
			self.dataset.signal_selection_changed.disconnect(callback=callback)
		self._bqplot["cleanups"].append(cleanup)

		return plot


	def plot_bq(self, grid=None, size=256, limits=None, square=False, center=None, weight=None, figsize=None,
			 aspect="auto", f="identity", fig=None, axes=None, xlabel=None, ylabel=None, title=None,
			 group_by=None, group_limits=None, group_colors='jet', group_labels=None, group_count=None,
			 cmap="afmhot", scales=None, tool_select=False, bq_cleanup=True,
			 **kwargs):
		import vaex.ext.bqplot
		import bqplot.interacts
		import bqplot.pyplot as p
		import ipywidgets as widgets
		import bqplot as bq
		f = _parse_f(f)
		limits = self.limits(limits)
		import vaex.ext.bqplot
		vaex.ext.bqplot.patch()
		if not hasattr(self, "_bqplot"):
			self._bqplot = {}
			self._bqplot["cleanups"] = []
		else:
			if bq_cleanup:
				for cleanup in self._bqplot["cleanups"]:
					cleanup()
			self._bqplot["cleanups"] = []
		if limits is None:
			limits = self.limits_sigma()
		#if fig is None:
		if scales is None:
			x_scale = bq.LinearScale(min=limits[0][0], max=limits[0][1])
			y_scale = bq.LinearScale(min=limits[1][0], max=limits[1][1])
			scales = {'x': x_scale, 'y': y_scale}
		else:
			x_scale = scales["x"]
			y_scale = scales["y"]
		if 1:
			fig = p.figure() # actually, bqplot doesn't return it
			fig = p.current_figure()
			fig.fig_color = "black" # TODO, take the color from the colormap
			fig.padding_y = 0
			# if we don't do this, bqplot may flip some axes... report this bug
			x = np.arange(10)
			y = x**2
			p.plot(x, y, scales=scales)
			#p.xlim(*limits[0])
			#p.ylim(*limits[1])
			#if grid is None:
		if group_limits is None and group_by:
			group_limits = tuple(self.dataset(group_by).minmax()[0]) + (group_count,)
		#fig = p.
		#if xlabel:
		fig.axes[0].label = xlabel or self.expressions[0]
		#if ylabel:
		fig.axes[1].label = ylabel or self.expressions[1]
		if title:
			fig.title = title
		#axes.set_aspect(aspect)
		rgba8 = self.image_rgba(grid=grid, size=size, limits=limits, square=square, center=center, weight=weight,
			 f=f, axes=axes,
			 group_by=group_by, group_limits=group_limits, group_colors=group_colors, group_count=group_count,
			 cmap=cmap)
		#x_scale = p._context["scales"]["x"]
		#y_scale = p._context["scales"]["y"]
		src="http://localhost:8888/kernelspecs/python2/logo-64x64.png"
		import bqplot.marks
		im = vaex.ext.bqplot.Image(src=src, scales=scales, x=0, y=0, width=1, height=1)
		if 0:
			size = 20
			x_data = np.arange(size)
			line = bq.Lines(x=x_data, y=np.random.randn(size), scales={'x': x_scale, 'y': y_scale},
							stroke_width=3, colors=['red'])


			ax_x = bq.Axis(scale=x_scale, tick_format='0.2f', grid_lines='solid')
			ax_y = bq.Axis(scale=y_scale, orientation='vertical', tick_format='0.2f', grid_lines='solid')
			panzoom = bq.PanZoom(scales={'x': [x_scale], 'y': [y_scale]})
			lasso = bqplot.interacts.LassoSelector()
			brush = bqplot.interacts.BrushSelector(x_scale=x_scale, y_scale=y_scale, color="green")
			fig = bq.Figure(marks=[line,im], axes=[ax_x, ax_y], min_width=100, min_height=100, interaction=panzoom)
		else:
			fig.marks = list(fig.marks) + [im]
		def make_image(executor, limits):
			#print "make image" * 100
			self.executor = executor
			if self.dataset.has_selection():
				sub = self.selected()
			else:
				sub = self
			return sub.image_rgba(limits=limits, size=size, f=f)
		progress = widgets.FloatProgress(value=0.0, min=0.0, max=1.0, step=0.01)
		updater = vaex.ext.bqplot.DebouncedThreadedUpdater(self, size, im, make_image, progress_widget=progress)
		def update_image():
			limits = [x_scale.min, x_scale.max], [y_scale.min, y_scale.max]
			#\print limits
			#print "update...", limits
			#vxbq.debounced_threaded_update(self.dataset, im, make_image2, limits=limits)
			updater.update(limits)
		def update(*args):
			update_image()
		y_scale.observe(update, "min")
		y_scale.observe(update, "max")
		x_scale.observe(update, "min")
		x_scale.observe(update, "max")
		update_image()
		#fig = kwargs.pop('figure', p.current_figure())
		tools = []
		tool_actions = []
		panzoom = bq.PanZoom(scales={'x': [x_scale], 'y': [y_scale]})
		tool_actions_map = {u"m":panzoom}
		tool_actions.append(u"m")

		fig.interaction = panzoom
		if tool_select:
			brush = bqplot.interacts.BrushSelector(x_scale=x_scale, y_scale=y_scale, color="green")
			tool_actions_map["b"] = brush
			tool_actions.append("b")
			def update_selection(*args):
				def f():
					if brush.selected:
						(x1, y1), (x2, y2) = brush.selected
						ex1, ex2 = self.expressions
						mode = modes_names[modes_labels.index(button_selection_mode.value)]
						self.dataset.select_rectangle(ex1, ex2, limits=[[x1, x2], [y1, y2]], mode=mode)
					else:
						self.dataset.select_nothing()
				updater.update_select(f)
			brush.observe(update_selection, "selected")
			#fig.interaction = brush
			#callback = self.dataset.signal_selection_changed.connect(lambda dataset: update_image())
			callback = self.dataset.signal_selection_changed.connect(lambda dataset: updater.update_direct_safe())
			def cleanup(callback=callback):
				self.dataset.signal_selection_changed.disconnect(callback=callback)
			self._bqplot["cleanups"].append(cleanup)

			button_select_nothing = widgets.Button(icon="fa-trash-o")
			def select_nothing(button):
				self.dataset.select_nothing()
			button_select_nothing.on_click(select_nothing)
			tools.append(button_select_nothing)
			modes_names = "replace and or xor subtract".split()
			modes_labels = "= & | ^ -".split()
			button_selection_mode = widgets.ToggleButtons(description='',options=modes_labels)
			tools.append(button_selection_mode)
		def change_interact(*args):
			#print "change", args
			fig.interaction = tool_actions_map[button_action.value]
		#tool_actions = ["m", "b"]
		#tool_actions = [("m", "m"), ("b", "b")]
		button_action = widgets.ToggleButtons(description='',options=tool_actions, icons=["fa-arrows", "fa-pencil-square-o"])
		button_action.observe(change_interact, "value")
		tools.insert(0,button_action)
		button_action.value = "m" #tool_actions[-1]
		if len(tools) == 1:
			tools = []
		tools = widgets.HBox(tools)

		box_layout = widgets.Layout(display='flex',
						flex_flow='column',
						#border='solid',
						width='100%', height="100%")
		fig.fig_margin = {'bottom': 40, 'left': 60, 'right': 10, 'top': 40}
		#fig.min_height = 700
		#fig.min_width = 400
		fig.layout = box_layout
		return widgets.VBox([fig, progress, tools])

	def figlarge(self, size=(10,10)):
		import pylab
		pylab.figure(num=None, figsize=size, dpi=80, facecolor='w', edgecolor='k')

	#def bounded(self):
	#	return self.bounded_by_minmax()

	def bounded_by(self, limits):
		"""Returns a bounded subspace (SubspaceBounded) with limits as given by limits

		:param limits: sequence of [(min, max), ..., (min, max)] values
		:rtype: SubspaceBounded
		"""
		return SubspaceBounded(self, np.array(limits))

	def bounded_by_minmax(self):
		"""Returns a bounded subspace (SubspaceBounded) with limits given by Subspace.minmax()

		:rtype: SubspaceBounded
		"""
		bounds = self.minmax()
		return SubspaceBounded(self, bounds)

	bounded = bounded_by_minmax

	def bounded_by_sigmas(self, sigmas=3, square=False):
		"""Returns a bounded subspace (SubspaceBounded) with limits given by Subspace.limits_sigma()

		:rtype: SubspaceBounded
		"""
		bounds = self.limits_sigma(sigmas=sigmas, square=square)
		return SubspaceBounded(self, bounds)

	def minmax(self):
		"""Return a sequence of [(min, max), ..., (min, max)] corresponding to each expression in this subspace ignoring NaN.
		"""
		raise NotImplementedError
	def mean(self):
		"""Return a sequence of [mean, ... , mean] corresponding to the mean of each expression in this subspace ignoring NaN.
		"""
		raise NotImplementedError
	def var(self, means=None):
		"""Return a sequence of [var, ... , var] corresponding to the variance of each expression in this subspace ignoring NaN.
		"""
		raise NotImplementedError
	def sum(self):
		"""Return a sequence of [sum, ... , sum] corresponding to the sum of values of each expression in this subspace ignoring NaN."""
		raise NotImplementedError
	def histogram(self, limits, size=256, weight=None):
		"""Return a grid of shape (size, ..., size) corresponding to the dimensionality of this subspace containing the counts in each element

		The type of the grid of np.float64

		"""
		raise NotImplementedError
	def limits_sigma(self, sigmas=3, square=False):
		raise NotImplementedError

	def row(self, index):
		return np.array([self.dataset.evaluate(expression, i1=index, i2=index+1)[0] for expression in self.expressions])


class SubspaceLocal(Subspace):
	"""Subclass of subspace which implemented methods that can be run locally.
	"""

	def _toarray(self, list):
		return np.array(list)

	@property
	def pre(self):
		self.executor.pre

	@property
	def post(self):
		self.executor.post

	def _task(self, task, progressbar=False):
		"""Helper function for returning tasks results, result when immediate is True, otherwise the task itself, which is a promise"""
		if self.delay:
			# should return a task or a promise nesting it
			return self.executor.schedule(task)
		else:
			import vaex.utils
			callback = None
			try:
				if progressbar == True:
					def update(fraction):
						bar.update(fraction)
						return True
					bar = vaex.utils.progressbar(task.name)
					callback = self.executor.signal_progress.connect(update)
				elif progressbar:
					callback = self.executor.signal_progress.connect(progressbar)
				result = self.executor.run(task)
				if progressbar == True:
					bar.finish()
					sys.stdout.write('\n')
				return result
			finally:
				if callback:
					self.executor.signal_progress.disconnect(callback)

	def minmax(self, progressbar=False):
		def min_max_reduce(minmax1, minmax2):
			if minmax1 is None:
				return minmax2
			if minmax2 is None:
				return minmax1
			result = []
			for d in range(self.dimension):
				min1, max1 = minmax1[d]
				min2, max2 = minmax2[d]
				result.append((min(min1, min2), max(max1, max2)))
			return result
		def min_max_map(thread_index, i1, i2, *blocks):
			if self.is_masked or self.dataset.filtered:
				mask = self.dataset.evaluate_selection_mask("default" if self.is_masked else None, i1=i1, i2=i2)
				blocks = [block[mask] for block in blocks]
				is_empty = all(~mask)
				if is_empty:
					return None
			#with lock:
			#print blocks
			#with lock:
			#	print thread_index, i1, i2, blocks
			blocks = [_asfloat(block) for block in blocks]
			return [vaex.vaexfast.find_nan_min_max(block) for block in blocks]
			if 0: # TODO: implement using statisticNd and benchmark
				minmaxes = np.zeros((len(blocks), 2), dtype=float)
				minmaxes[:,0] = np.inf
				minmaxes[:,1] = -np.inf
				for i, block in enumerate(blocks):
					vaex.vaexfast.statisticNd([], block, minmaxes[i,:], [], [], 2)
				#minmaxes[~np.isfinite(minmaxes)] = np.nan
				return minmaxes
		task = TaskMapReduce(self.dataset, self.expressions, min_max_map, min_max_reduce, self._toarray, info=True, name="minmax")
		return self._task(task, progressbar=progressbar)

	def mean(self):
		return self._moment(1)

	def _moment(self, moment=1):
		def mean_reduce(means_and_counts1, means_and_counts2):
			means_and_counts = []
			for (mean1, count1), (mean2, count2) in zip(means_and_counts1, means_and_counts2):
				means_and_counts.append( [np.nansum([mean1*count1, mean2*count2])/(count1+count2), count1+count2] )
			return means_and_counts
		def remove_counts(means_and_counts):
			return self._toarray(means_and_counts)[:,0]
		def mean_map(thread_index, i1, i2, *blocks):
			if self.is_masked or self.dataset.filtered:
				mask = self.dataset.evaluate_selection_mask("default" if self.is_masked else None, i1=i1, i2=i2)
				return [(np.nanmean(block[mask]**moment), np.count_nonzero(~np.isnan(block[mask]))) for block in blocks]
			else:
				return [(np.nanmean(block**moment), np.count_nonzero(~np.isnan(block))) for block in blocks]
		task = TaskMapReduce(self.dataset, self.expressions, mean_map, mean_reduce, remove_counts, info=True)
		return self._task(task)

	def var(self, means=None):
		# variances are linear, use the mean to reduce
		def vars_reduce(vars_and_counts1, vars_and_counts2):
			vars_and_counts = []
			for (var1, count1), (var2, count2) in zip(vars_and_counts1, vars_and_counts2):
				vars_and_counts.append( [np.nansum([var1*count1, var2*count2])/(count1+count2), count1+count2] )
			return vars_and_counts
		def remove_counts(vars_and_counts):
			return self._toarray(vars_and_counts)[:,0]
		if self.is_masked or self.dataset.filtered:
			def var_map(thread_index, i1, i2, *blocks):
				mask = self.dataset.evaluate_selection_mask("default" if self.is_masked else None, i1=i1, i2=i2)
				if means is not None:
					return [(np.nanmean((block[mask]-mean)**2), np.count_nonzero(~np.isnan(block[mask]))) for block, mean in zip(blocks, means)]
				else:
					return [(np.nanmean(block[mask]**2), np.count_nonzero(~np.isnan(block[mask]))) for block in blocks]
			task = TaskMapReduce(self.dataset, self.expressions, var_map, vars_reduce, remove_counts, info=True)
		else:
			def var_map(*blocks):
				if means is not None:
					return [(np.nanmean((block-mean)**2), np.count_nonzero(~np.isnan(block))) for block, mean in zip(blocks, means)]
				else:
					return [(np.nanmean(block**2), np.count_nonzero(~np.isnan(block))) for block in blocks]
			task = TaskMapReduce(self.dataset, self.expressions, var_map, vars_reduce, remove_counts)
		return self._task(task)

	def correlation(self, means=None, vars=None):
		if self.dimension != 2:
			raise ValueError("correlation is only defined for 2d subspaces, not %dd" % self.dimension)

		def do_correlation(means, vars):
			meanx, meany = means
			sigmax, sigmay = vars[0]**0.5, vars[1]**0.5

			def remove_counts_and_normalize(covar_and_count):
				covar, counts = covar_and_count
				return covar/counts / (sigmax * sigmay)

			def covars_reduce(covar_and_count1, covar_and_count2):
				if covar_and_count1 is None:
					return covar_and_count2
				if covar_and_count2 is None:
					return covar_and_count1
				else:
					covar1, count1 = covar_and_count1
					covar2, count2 = covar_and_count2
					return [np.nansum([covar1, covar2]), count1+count2]

			mask = self.dataset.mask
			def covar_map(thread_index, i1, i2, *blocks):
				#return [(np.nanmean((block[mask[i1:i2]]-mean)**2), np.count_nonzero(~np.isnan(block[mask[i1:i2]]))) for block, mean in zip(blocks, means)]
				blockx, blocky = blocks
				if self.is_masked:
					blockx, blocky = blockx[mask[i1:i2]], blocky[mask[i1:i2]]
				counts = np.count_nonzero( ~(np.isnan(blockx) | np.isnan(blocky)) )
				if counts == 0:
					return None
				else:
					return np.nansum((blockx - meanx) * (blocky - meany)), counts

			task = TaskMapReduce(self.dataset, self.expressions, covar_map, covars_reduce, remove_counts_and_normalize, info=True)
			return self._task(task)
		if means is None:
			if self.delay:
				means_wrapper = [None]
				def do_vars(means):
					means_wrapper[0] = means
					return self.var(means)
				def do_correlation_wrapper(vars):
					return do_correlation(means_wrapper[0], vars)
				return self.mean().then(do_vars).then(do_correlation_wrapper)
			else:
				means = self.mean()
				vars = self.var(means=means)
				return do_correlation(means, vars)
		else:
			if vars is None:
				if self.delay:
					def do_correlation_wrapper(vars):
						return do_correlation(means, vars)
					return self.vars(means=means).then(do_correlation_wrapper)
				else:
					vars = self.var(means)
					return do_correlation(means, vars)
			else:
				if means is None:
					means = self.mean()
				if vars is None:
					vars = self.var(means=means)
				return do_correlation(means, vars)

	def sum(self):
		nansum = lambda x: np.nansum(x, dtype=np.float64)
		# TODO: we can speed up significantly using our own nansum, probably the same for var and mean
		nansum = vaex.vaexfast.nansum
		if self.is_masked or self.dataset.filtered:
			task = TaskMapReduce(self.dataset,\
								 self.expressions, lambda thread_index, i1, i2, *blocks: [nansum(block[self.dataset.evaluate_selection_mask("default" if self.is_masked else None, i1=i1, i2=i2)])
																						  for block in blocks],\
								 lambda a, b: np.array(a) + np.array(b), self._toarray, info=True)
		else:
			task = TaskMapReduce(self.dataset, self.expressions, lambda *blocks: [nansum(block) for block in blocks], lambda a, b: np.array(a) + np.array(b), self._toarray)
		return self._task(task)

	def histogram(self, limits, size=256, weight=None, progressbar=False, group_by=None, group_limits=None):
		expressions = self.expressions
		if group_by:
			expressions = list(expressions) + [group_by]
			limits = list(limits) + [group_limits[:2]] #[[group_limits[0] - 0,5, group_limits[1]+0.5]]
			#assert group_limits[2] == 1
			size = (group_limits[2],) + (size,) * (len(expressions) -1)
		task = TaskHistogram(self.dataset, self, expressions, size, limits, masked=self.is_masked, weight=weight)
		return self._task(task, progressbar=progressbar)

	def bin_mean(self, expression, limits, size=256, progressbar=False, group_by=None, group_limits=None):
		# todo, fix progressbar into two...
		counts = self.histogram(limits=limits, size=size, progressbar=progressbar, group_by=group_by, group_limits=group_limits)
		weighted  =self.histogram(limits=limits, size=size, progressbar=progressbar, group_by=group_by, group_limits=group_limits,
								weight=expression)
		mean = weighted/counts
		mean[counts==0] = np.nan
		return mean

	def bin_mean_cyclic(self, expression, max_value, limits, size=256, progressbar=False, group_by=None, group_limits=None):
		# todo, fix progressbar into two...
		meanx = self.bin_mean(limits=limits, size=size, progressbar=progressbar, group_by=group_by, group_limits=group_limits,
								expression="cos((%s)/%r*2*pi)" % (expression, max_value))
		meany = self.bin_mean(limits=limits, size=size, progressbar=progressbar, group_by=group_by, group_limits=group_limits,
								expression="sin((%s)/%r*2*pi)" % (expression, max_value))
		angles = np.arctan2(meany, meanx)
		values =  ((angles+2*np.pi) % (2*np.pi)) / (2*np.pi) * max_value
		length =  np.sqrt(meanx**2+meany**2)
		length[~np.isfinite(meanx)] = np.nan
		return values, length

	def mutual_information(self, limits=None, grid=None, size=256):
		if limits is None:
			limits_done = Task.fulfilled(self.minmax())
		else:
			limits_done = Task.fulfilled(limits)
		if grid is None:
			if limits is None:
				histogram_done = limits_done.then(lambda limits: self.histogram(limits, size=size))
			else:
				histogram_done = Task.fulfilled(self.histogram(limits, size=size))
		else:
			histogram_done = Task.fulfilled(grid)
		mutual_information_promise = histogram_done.then(vaex.kld.mutual_information)
		return mutual_information_promise if self.delay else mutual_information_promise.get()


	def limits_percentage(self, percentage=99.73, square=False):
		import scipy.ndimage
		limits = []
		for expr in self.expressions:
			subspace = self.dataset(expr)
			if self.is_selected():
				subspace = subspace.selected()
			limits_minmax = subspace.minmax()
			vmin, vmax = limits_minmax[0]
			size = 1024*16
			counts = subspace.histogram(size=size, limits=limits_minmax)
			cumcounts = np.concatenate([[0], np.cumsum(counts)])
			cumcounts /= cumcounts.max()
			# TODO: this is crude.. see the details!
			f = (1-percentage/100.)/2
			x = np.linspace(vmin, vmax, size+1)
			l = scipy.interp([f,1-f], cumcounts, x)
			limits.append(l)
		return limits

	def limits_sigma(self, sigmas=3, square=False):
		if self.delay:
			means_wrapper = [None]
			def do_vars(means):
				means_wrapper[0] = means
				return self.var(means)
			def do_limits(vars):
				stds = vars**0.5
				means = means_wrapper[0]
				if square:
					stds = np.repeat(stds.mean(), len(stds))
				return np.array(list(zip(means-sigmas*stds, means+sigmas*stds)))
			return self.mean().then(do_vars).then(do_limits)
		else:
			means = self.mean()
			stds = self.var(means=means)**0.5
			if square:
				stds = np.repeat(stds.mean(), len(stds))
			return np.array(list(zip(means-sigmas*stds, means+sigmas*stds)))

	def _not_needed_current(self):
		index = self.dataset.get_current_row()
		def find(thread_index, i1, i2, *blocks):
			if (index >= i1) and (index < i2):
				return [block[index-i1] for block in blocks]
			else:
				return None
		task = TaskMapReduce(self.dataset, self.expressions, find, lambda a, b: a if b is None else b, info=True)
		return self._task(task)

	def nearest(self, point, metric=None):
		metric = metric or [1.] * len(point)
		def nearest_in_block(thread_index, i1, i2, *blocks):
			if self.is_masked:
				mask = self.dataset.evaluate_selection_mask("default", i1=i1, i2=i2)
				if mask.sum() == 0:
					return None
				blocks = [block[mask] for block in blocks]
			distance_squared = np.sum( [(blocks[i]-point[i])**2.*metric[i] for i in range(self.dimension)], axis=0 )
			min_index_global = min_index = np.argmin(distance_squared)
			if self.is_masked: # we skipped some indices, so correct for that
				min_index_global = np.argmin((np.cumsum(mask) - 1 - min_index)**2)
			#with lock:
			#	print i1, i2, min_index, distance_squared, [block[min_index] for block in blocks]
			return min_index_global.item() + i1, distance_squared[min_index].item()**0.5, [block[min_index].item() for block in blocks]
		def nearest_reduce(a, b):
			if a is None:
				return b
			if b is None:
				return a
			if a[1] < b[1]:
				return a
			else:
				return b
		if self.is_masked:
			pass
		task = TaskMapReduce(self.dataset,\
							 self.expressions,
							 nearest_in_block,\
							 nearest_reduce, info=True)
		return self._task(task)






__author__ = 'breddels'
import logging
logger = logging.getLogger("vaex.file")

import vaex.file.other
try:
	import vaex.hdf5 as hdf5
except ImportError:
	hdf5 = None
if hdf5:
	import vaex.hdf5.dataset

def can_open(path, *args, **kwargs):
	for name, class_ in list(vaex.file.other.dataset_type_map.items()):
		if class_.can_open(path, *args):
			return True


def open(path, *args, **kwargs):
	dataset_class = None
	openers = []
	if hdf5:
		openers.extend(hdf5.dataset.dataset_type_map.items())
	openers.extend(vaex.file.other.dataset_type_map.items())
	for name, class_ in list(openers):
		logger.debug("trying %r with class %r" % (path, class_))
		if class_.can_open(path, *args, **kwargs):
			logger.debug("can open!")
			dataset_class = class_
			break
	if dataset_class:
		dataset = dataset_class(path, *args, **kwargs)
		return dataset



from __future__ import print_function
__author__ = 'maartenbreddels'
import vaex
import vaex.utils
import logging
import astropy.units
import sys
import yaml

logger = logging.getLogger("vaex.meta")

def main(argv):
	import argparse
	parser = argparse.ArgumentParser(argv[0])
	parser.add_argument('--verbose', '-v', action='count', default=0)
	parser.add_argument('--list', '-l', default=False, action='store_true', help="list columns of input")

	subparsers = parser.add_subparsers(help='type of subtask', dest="task")

	parser_export = subparsers.add_parser('export', help='read meta info')
	parser_export.add_argument("input", help="input dataset")
	parser_export.add_argument('output', help='output file (.yaml or .json)')
	parser_export.add_argument("columns", help="list of columns to export (or all when empty)", nargs="*")
	parser_export.add_argument('--all', dest="all", action='store_true', default=False, help="Also export missing values (useful for having a template)")

	parser_import = subparsers.add_parser('import', help='read meta info')
	parser_import.add_argument('input', help='input meta file (.yaml or .json)')
	parser_import.add_argument("output", help="output dataset")
	parser_import.add_argument("columns", help="list of columns to export (or all when empty)", nargs="*")
	parser_import.add_argument('--overwrite', help="overwrite existing entries", default=False, action='store_true')

	args = parser.parse_args(argv[1:])

	verbosity = ["ERROR", "WARNING", "INFO", "DEBUG"]
	logging.getLogger("vaex").setLevel(verbosity[min(3, args.verbose)])

	if args.task == "export":
		ds = vaex.open(args.input)
		column_names = ds.get_column_names(strings=True, virtual=True)
		if args.all:
			output_data = dict(description=ds.description,
						 	descriptions={name:ds.descriptions.get(name, "") for name in column_names},
							ucds={name:ds.ucds.get(name, "") for name in column_names},
							units={name:str(ds.units.get(name, "")) for name in column_names}, #{name:str(unit) for name, unit in ds.units.items()},
						   )
		else:
			output_data = dict(description=ds.description,
						 	descriptions=ds.descriptions,
							ucds=ds.ucds,
							units={name:str(unit) for name, unit in ds.units.items()},
						   )
		if args.output == "-":
			yaml.safe_dump(output_data, sys.stdout, default_flow_style=False) #, encoding='utf-8',  allow_unicode=True)
		else:
			vaex.utils.write_json_or_yaml(args.output, output_data)
			print("wrote %s" % args.output)
	if args.task == "import":
		if args.input == "-":
			data = yaml.load(sys.stdin)
		else:
			data = vaex.utils.read_json_or_yaml(args.input)

		ds = vaex.open(args.output)

		units = data["units"]
		ucds = data["ucds"]
		for column_name in ds.get_column_names():
			if (args.overwrite or column_name not in ds.units) and column_name in units:
				ds.units[column_name] = astropy.units.Unit(units[column_name])
			if (args.overwrite or column_name not in ds.ucds) and column_name in ucds:
				ds.ucds[column_name] = ucds[column_name]
		ds.write_meta()
		print("updated meta data in %s" % args.output)

if __name__ == "__main__":
	main(sys.argv)




from .expression import _unary_ops, _binary_ops, reversable
from future.utils import with_metaclass
from vaex.dataset import expression_namespace
from vaex.delayed import delayed

class Meta(type):
    def __new__(upperattr_metaclass, future_class_name,
                future_class_parents, attrs):
        for op in _binary_ops:
            def wrap(op=op):
                def f(a, b):
                    return _StatisticsCalculation(op['name'], op['op'], [a, b], binary=True, code=op['code'])
                attrs['__%s__' % op['name']] = f
                if op['name'] in reversable:
                    def f(a, b):
                        return _StatisticsCalculation(op['name'], op['op'], [b, a], binary=True, code=op['code'])
                    attrs['__r%s__' % op['name']] = f
            wrap(op)
        for op in _unary_ops:
            def wrap(op=op):
                def f(a):
                    return _StatisticsCalculation(op['name'], op['op'], [a], unary=True, code=op['code'])
                attrs['__%s__' % op['name']] = f
            wrap(op)
        for name, func_real in expression_namespace.items():
            def wrap(name=name, func_real=func_real):
                def f(*args, **kwargs):
                    return _StatisticsCalculation(name, func_real, args)
                attrs['%s' % name] = f
            if name not in attrs:
                wrap(name)
        return type(future_class_name, future_class_parents, attrs)

class Expression(with_metaclass(Meta)):
    '''Describes an expression for a statistic'''
    def calculate(self, ds, binby=[], shape=256, limits=None, selection=None):
        '''Calculate the statistic for a :class:`Dataset`'''
        raise NotImplementedError()

    def __repr__(self):
        return '{}'.format(self)

class _StatisticsCalculation(Expression):
    def __init__(self, name, op, args, binary=False, unary=False, code='"<??>"'):
        self.name = name
        self.op = op
        self.args = args
        self.binary = binary
        self.unary = unary
        self.code = code

    def __str__(self):
        if self.binary:
            return "({0} {1} {2})".format(repr(self.args[0]), self.code, repr(self.args[1]))
        if self.unary:
            return "{0}{1}".format(self.code, repr(self.args[0]))
        return "{0}({1})".format(self.name, ", ".join(repr(k) for k in self.args))

    def calculate(self, ds, binby=[], shape=256, limits=None, selection=None, delay=False):
        def to_value(v):
            if isinstance(v, Expression):
                return v.calculate(ds, binby=binby, shape=shape, limits=limits, selection=selection, delay=delay)
            return v
        values = [to_value(v) for v in self.args]
        #print(values, self.op)
        op = self.op
        if delay:
            op = delayed(op)
        return op(*values)


class _Statistic(Expression):
    def __init__(self, name, *expression):
        self.name = name
        self.expression = expression
        self.args = self.expression

    def __str__(self):
        return "{0}({1})".format(self.name, ", ".join(str(k) for k in self.args))

    def calculate(self, ds, binby=[], shape=256, limits=None, selection=None, delay=False):
        method = getattr(ds, self.name)
        return method(*self.args, binby=binby, shape=shape, limits=limits, selection=selection, delay=delay)

def count(expression='*'):
    '''Creates a count statistic'''  
    return _Statistic('count', expression)

def sum(expression):
    '''Creates a sum statistic'''
    return _Statistic('sum', expression)

def mean(expression):
    '''Creates a mean statistic'''
    return _Statistic('mean', expression)

def std(expression):
   '''Creates a standard deviation statistic'''
   return _Statistic('std', expression)

def covar(x, y):
   '''Creates a standard deviation statistic'''
   return _Statistic('covar', x, y)

def correlation(x, y):
   '''Creates a standard deviation statistic'''
   return _Statistic('correlation', x, y)



import traceback
import logging
logger = logging.getLogger("vaex.events")

class Signal(object):
	def __init__(self, name=None):
		"""

		:type name: str
		:return:
		"""
		self.name = name or repr(self)
		self.callbacks = []
		self.extra_args = {}
		
	def connect(self, callback, prepend=False, *args, **kwargs):
		logger.debug("(%s) connected %s", self.name, callback)
		if prepend:
			self.callbacks.insert(0, callback)
		else:
			self.callbacks.append(callback)
		self.extra_args[callback] = (args, kwargs)
		return callback
						   
	def emit(self, *args, **kwargs):
		results = []
		for callback in list(self.callbacks): # copy it because handlers van add or remove items
			extra_args, extra_kwargs = self.extra_args[callback]
			final_args = args + extra_args
			final_kwargs = {}
			final_kwargs.update(extra_kwargs)
			final_kwargs.update(kwargs)
			try:
				logger.debug("(%s) calling %r with arguments %r and kwargs %r" % (self.name, callback, final_args, final_kwargs))
				value = callback(*final_args, **final_kwargs)
				results.append(value)
			except Exception:
				logger.error("error in handling callback %r with arguments %r and kwargs %r" % (callback, final_args, final_kwargs))
				raise
				#tb = traceback.format_exc()
				#raise Exception("error while calling callback: %r with arguments %r and kwargs %r" % (callback, final_args, final_kwargs), tb)
		return results
			
	def disconnect(self, callback):
		self.callbacks.remove(callback)
		del self.extra_args[callback]
		


# -*- coding: utf-8 -*-
from __future__ import division, print_function
import threading
import queue
import math
import multiprocessing
import time
import sys
import vaex.utils
import logging
import traceback
import cProfile
import concurrent.futures
import time


thread_count_default = multiprocessing.cpu_count()# * 2 + 1
logger = logging.getLogger("vaex.multithreading")

class ThreadPoolIndex(concurrent.futures.ThreadPoolExecutor):
	"""Thread pools that adds a thread index as first argument to the callback passed to map

	This is useful if you keep a piece of memory (like ndgrid) per array
	"""
	def __init__(self, max_workers=thread_count_default, *args, **kwargs):
		super(ThreadPoolIndex, self).__init__(max_workers, *args, **kwargs)
		self.lock = threading.Lock()
		self.thread_indices = iter(range(1000000)) # enough threads until 2100?
		self.local = threading.local()
		self.nthreads = self._max_workers

	def map(self, callable, iterator, on_error=None, progress=None, cancel=None, unpack=False):
		progress = progress or (lambda x: True)
		cancelled = False
		def wrapped(*args, **kwargs):
			if not cancelled:
				with self.lock:
					if not hasattr(self.local, 'index'):
						self.local.index = next(self.thread_indices)
				if unpack:
					args = args[0] # it's passed as a tuple.. not sure why
				callable(self.local.index, *args, **kwargs)
		# convert to list so we can count
		values = list(iterator)
		N = len(values)
		time_last = time.time()-100
		min_delta_t = 1./100 # max 100 per second
		for i, value in enumerate(super(ThreadPoolIndex, self).map(wrapped, values)):
			progress_value = (i+1)/N
			time_now = time.time()
			if progress_value == 1 or (time_now - time_last) > min_delta_t:
				time_last = time_now
				if progress(progress_value) == False:
					cancelled = True
			yield value



main_pool = None #ThreadPoolIndex()
def get_main_pool():
	global main_pool
	if main_pool is None:
		main_pool = ThreadPoolIndex()
	return main_pool




import re
import string
import sys
import time


PY2 = sys.version_info[0] == 2
if not PY2:
  text_type = str
  string_types = (str,)
  unichr = chr
else:
  text_type = unicode
  string_types = (str, unicode)
  unichr = unichr


COMMENT = object()


def load_properties(fh, mapping=dict):
  """
    Reads properties from a Java .properties file.

    Returns a dict (or provided mapping) of properties.

    :param fh: a readable file-like object
    :param mapping: mapping type to load properties into
  """
  return mapping(iter_properties(fh))


def store_properties(fh, props, comment=None, timestamp=True):
  """
    Writes properties to the file in Java properties format.

    :param fh: a writable file-like object
    :param props: a mapping (dict) or iterable of key/value pairs
    :param comment: comment to write to the beginning of the file
    :param timestamp: boolean indicating whether to write a timestamp comment
  """
  if comment is not None:
    write_comment(fh, comment)

  if timestamp:
    write_comment(fh, time.strftime('%a %b %d %H:%M:%S %Z %Y'))

  if hasattr(props, 'keys'):
    for key in props:
      write_property(fh, key, props[key])
  else:
    for key, value in props:
      write_property(fh, key, value)


def write_comment(fh, comment):
  """
    Writes a comment to the file in Java properties format.

    Newlines in the comment text are automatically turned into a continuation
    of the comment by adding a "#" to the beginning of each line.

    :param fh: a writable file-like object
    :param comment: comment string to write
  """
  _require_string(comment, 'comments')
  fh.write(_escape_comment(comment))
  fh.write(b'\n')


def _require_string(value, name):
  if isinstance(value, string_types):
    return

  valid_types = ' or '.join(cls.__name__ for cls in string_types)
  raise TypeError('%s must be %s, but got: %s %r'
                  % (name, valid_types, type(value), value))


def write_property(fh, key, value):
  """
    Write a single property to the file in Java properties format.

    :param fh: a writable file-like object
    :param key: the key to write
    :param value: the value to write
  """
  if key is COMMENT:
    write_comment(fh, value)
    return

  _require_string(key, 'keys')
  _require_string(value, 'values')

  fh.write(_escape_key(key))
  fh.write(b'=')
  fh.write(_escape_value(value))
  fh.write(b'\n')


def iter_properties(fh, comments=False):
  """
    Incrementally read properties from a Java .properties file.

    Yields tuples of key/value pairs.

    If ``comments`` is `True`, comments will be included with ``jprops.COMMENT``
    in place of the key.

    :param fh: a readable file-like object
    :param comments: should include comments (default: False)
  """
  for line in _property_lines(fh):
    key, value = _split_key_value(line)
    if key is not COMMENT:
      key = _unescape(key)
    elif not comments:
      continue
    yield key, _unescape(value)


################################################################################
# Helpers for property parsing/writing
################################################################################


_COMMENT_CHARS = '#!'
_COMMENT_CHARS_BYTES = bytearray(_COMMENT_CHARS, 'ascii')
_LINE_PATTERN = re.compile(br'^\s*(?P<body>.*?)(?P<backslashes>\\*)$')
_KEY_TERMINATORS_EXPLICIT = '=:'
_KEY_TERMINATORS = _KEY_TERMINATORS_EXPLICIT + string.whitespace

_KEY_TERMINATORS_EXPLICIT_BYTES = bytearray(_KEY_TERMINATORS_EXPLICIT, 'ascii')
_KEY_TERMINATORS_BYTES = bytearray(_KEY_TERMINATORS, 'ascii')


_escapes = {
  't': '\t',
  'n': '\n',
  'f': '\f',
  'r': '\r',
}
_escapes_rev = dict((v, '\\' + k) for k,v in _escapes.items())
for c in '\\' + _COMMENT_CHARS + _KEY_TERMINATORS_EXPLICIT:
  _escapes_rev.setdefault(c, '\\' + c)


def _unescape(value):
  # all values required to be latin-1 encoded bytes
  value = value.decode('latin-1')

  # if not native string (e.g. PY2) try converting it back
  if not isinstance(value, str):
    try:
      value = value.encode('ascii')
    except UnicodeEncodeError:
      # cannot be represented in ASCII so leave it as unicode type
      pass

  def unirepl(m):
    backslashes = m.group(1)
    charcode = m.group(2)

    # if preceded by even number of backslashes, the \u is escaped
    if len(backslashes) % 2 == 0:
      return m.group(0)

    c = unichr(int(charcode, 16))
    # if unicode decodes to '\', re-escape it to unescape in the second step
    if c == '\\':
      c = u'\\\\'

    return backslashes + c

  value = re.sub(r'(\\+)u([0-9a-fA-F]{4})', unirepl, value)

  def bslashrepl(m):
    code = m.group(1)
    return _escapes.get(code, code)

  return re.sub(r'\\(.)', bslashrepl, value)


def _escape_comment(comment):
  comment = comment.replace('\r\n', '\n').replace('\r', '\n')
  comment = re.sub(r'\n(?![#!])', '\n#', comment)
  if isinstance(comment, text_type):
    comment = re.sub(u'[\u0100-\uffff]', _unicode_replace, comment)
    comment = comment.encode('latin-1')
  return b'#' + comment


def _escape_key(key):
  return _escape(key, _KEY_TERMINATORS)


def _escape_value(value):
  tail = value.lstrip()
  if len(tail) == len(value):
    return _escape(value)

  if tail:
    head = value[:-len(tail)]
  else:
    head = value

  # escape any leading whitespace, but leave other spaces intact
  return _escape(head, string.whitespace) + _escape(tail)


def _escape(value, chars=''):
  escape_chars = set(_escapes_rev)
  escape_chars.update(chars)
  escape_pattern = '[%s]' % re.escape(''.join(escape_chars))

  def esc(m):
    c = m.group(0)
    return _escapes_rev.get(c) or '\\' + c
  value = re.sub(escape_pattern, esc, value)

  value = re.sub(u'[\u0000-\u0019\u007f-\uffff]', _unicode_replace, value)

  return value.encode('latin-1')


def _unicode_replace(m):
  c = m.group(0)
  return r'\u%.4x' % ord(c)


def _split_key_value(line):
  if line[0] in _COMMENT_CHARS_BYTES:
    return COMMENT, line[1:]

  escaped = False
  key_buf = bytearray()

  line_orig = line
  line = bytearray(line)

  for idx, c in enumerate(line):
    if not escaped and c in _KEY_TERMINATORS_BYTES:
      key_terminated_fully = c in _KEY_TERMINATORS_EXPLICIT_BYTES
      break

    key_buf.append(c)
    escaped = c == ord('\\')

  else:
    # no key terminator, key is full line & value is blank
    return line_orig, b''

  value = line[idx+1:].lstrip()
  if not key_terminated_fully and value[:1] in _KEY_TERMINATORS_EXPLICIT_BYTES:
    value = value[1:].lstrip()

  return bytes(key_buf), bytes(value)


def _universal_newlines(fp):
  """
    Wrap a file to convert newlines regardless of whether the file was opened
    with the "universal newlines" option or not.
  """
  # if file was opened with universal newline support we don't need to convert
  if 'U' in getattr(fp, 'mode', ''):
    for line in fp:
      yield line
  else:
    for line in fp:
      line = line.replace(b'\r\n', b'\n').replace(b'\r', b'\n')
      for piece in line.split(b'\n'):
        yield piece


def _property_lines(fp):
  buf = bytearray()
  for line in _universal_newlines(fp):
    m = _LINE_PATTERN.match(line)

    body = m.group('body')
    backslashes = m.group('backslashes')

    if len(backslashes) % 2 == 0:
      body += backslashes
      continuation = False
    else:
      body += backslashes[:-1]
      continuation = True

    if not body:
      continue

    buf.extend(body)

    if not continuation:
      yield bytes(buf)
      buf = bytearray()



from setuptools import setup
import sys, os, imp
from setuptools import Extension

dirname = os.path.dirname(__file__)
path_version = os.path.join(dirname, "vaex/core/_version.py")
version = imp.load_source('version', path_version)

name        = 'vaex'
author      = "Maarten A. Breddels"
author_email= "maartenbreddels@gmail.com"
license     = 'MIT'
version     = version.__version__
url         = 'https://www.github.com/maartenbreddels/vaex'
# TODO: can we do without requests and progressbar2?
# TODO: after python2 supports frops, future and futures can also be dropped
# TODO: would be nice to have astropy only as dep in vaex-astro
install_requires_core = ["numpy>=1.11", "astropy>=2", "aplus",
    "future>=0.15.2", "pyyaml", "progressbar2", "psutil>=1.2.1", "requests", "six", "cloudpickle"]
if sys.version_info[0] == 2:
    install_requires_core.append("futures>=2.2.0")
install_requires_viz = ["matplotlib>=1.3.1", ]
install_requires_astro = ["kapteyn"]

class get_numpy_include(object):
    """Helper class to determine the numpy include path
    The purpose of this class is to postpone importing numpy
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self):
        pass

    def __str__(self):
        import numpy as np
        return np.get_include()

extra_compile_args = ["-std=c++0x", "-mfpmath=sse", "-O3", "-funroll-loops"]

# on windows (Conda-forge builds), the dirname is an absolute path
extension_vaexfast = Extension("vaex.vaexfast", [os.path.relpath(os.path.join(dirname, "src/vaexfast.cpp"))],
                include_dirs=[get_numpy_include()],
                extra_compile_args=extra_compile_args)

setup(name=name+'-core',
    version=version,
    description='Core of vaex',
    url=url,
    author=author,
    author_email=author_email,
    setup_requires=['numpy'],
    install_requires=install_requires_core,
    license=license,
    package_data={'vaex': ['test/files/*.fits', 'test/files/*.vot', 'test/files/*.hdf5']},
    packages=['vaex', 'vaex.core', 'vaex.file', 'vaex.test', 'vaex.ext', 'vaex.misc'],
    ext_modules=[extension_vaexfast],
    zip_safe=False,
    entry_points={
        'console_scripts': ['vaex = vaex.__main__:main'],
        'gui_scripts': ['vaexgui = vaex.__main__:main'] # sometimes in osx, you need to run with this
        }
    )



__author__ = 'maartenbreddels'
import numpy as np
import vaex.vaexfast
from vaex.utils import filesize_format
import logging
import vaex.utils
total_bytes = 0

logger = logging.getLogger("vaex.grids")

def dog(grid, sigma1, sigma2):
	return gf(grid, sigma1) - gf(grid, sigma2)

def gf(grid, sigma, **kwargs):
	import scipy.ndimage
	return scipy.ndimage.gaussian_filter(grid, sigma=sigma, **kwargs)

functions = {}
functions["dog"] = dog
functions["gf"] = gf

def grid_average(scope, counts_name="counts", weighted_name="weighted"):

	counts = scope.evaluate(counts_name)
	weighted = scope.evaluate(weighted_name)
	logger.debug("evaluating average: counts=%r weighted=%r" % (counts, weighted))
	mask = counts==0
	average = np.zeros(counts.shape, dtype=np.float64)
	average[~mask] = weighted[~mask]/counts[~mask]
	average[mask] = np.nan
	return average

class GridScope(object):
	def __init__(self, locals=None, globals=None):
		#self.locals = locals or {}
		#if locals:
		#	self.__dict__.update(locals)
		self.globals = globals or {}
		self.lazy = {}
		self.lazy["average"] = grid_average
		self.globals["cumulative"] = self.cumulative
		self.globals["normalize"] = self.normalize
		self.globals.update(functions)
		self.user_added = set()

	def cumulative(self, array, normalize=True):
		mask = (np.isnan(array) | np.isinf(array))
		values = array * 1
		values[mask] = 0
		c = np.cumsum(values)
		if normalize:
			return c/c[-1]
		else:
			return c

	def normalize(self, array):
		mask = (np.isnan(array) | np.isinf(array))
		values = array * 1
		total = np.sum(values[~mask])
		return values/total

	def setter(self, key):
		def apply(value, key=key):
			self[key] = value
			return value
		return apply


	def add_lazy(self, key, f):
		self.lazy[key] = f

	def __contains__(self, item):
		return (item in self.__dict__) or (item in self.lazy)

	def __setitem__(self, key, value):
		#logger.debug("%r.__setitem__(%r, %r)" % (self, key, value))
		self.__dict__[key] = value
		self.user_added.add(key)

	def __getitem__(self, key):
		logger.debug("%r.__getitem__(%r)" % (self, key))
		if key in self.lazy and key not in self.__dict__:
			logger.debug("%r lazy item %r" % (self, key))
			f = self.lazy[key]
			self.__dict__[key] = f(self)

		return self.__dict__[key]

	def evaluate(self, expression):
		if 0:
			locals = dict(self.__dict__)
			del locals["globals"]
			logger.debug("evaluating: %r locals=%r", expression, locals)
		return eval(expression, self.globals, self)

	def slice(self, slice):
		gridscope = GridScope(globals=self.globals)
		for key in self.user_added:
			value = self[key]
			if isinstance(value, np.ndarray):
				grid = value
				sliced = np.sum(grid[slice,...], axis=0)
				logger.debug("sliced %s from %r to %r", key, grid.shape, sliced.shape)
				gridscope[key] = sliced
			else:
				gridscope[key] = value
		return gridscope

	def disjoined(self):
		gridscope = GridScope(globals=self.globals)
		for key in self.user_added:
			value = self[key]
			if isinstance(value, np.ndarray):
				grid = vaex.utils.disjoined(value)
				gridscope[key] = grid
			else:
				gridscope[key] = value
		return gridscope

	def marginal2d(self, i, j):
		gridscope = GridScope(globals=self.globals)
		for key in self.user_added:
			value = self[key]
			if isinstance(value, np.ndarray):
				dimension = len(value.shape)
				axes = list(range(dimension))
				axes.remove(i)
				axes.remove(j)
				grid = vaex.utils.multisum(value, axes)
				gridscope[key] = grid
			else:
				gridscope[key] = value
		return gridscope

def add_mem(bytes, *info):
	global total_bytes
	total_bytes += bytes
	added = filesize_format(bytes)
	total = filesize_format(total_bytes)
	print(("MEMORY USAGE: added %s, total %s (%r)" % (added, total, info)))




"""
Vaex is a library for dealing with big tabular data.

The most important class (datastructure) in vaex is the :class:`.Dataset`. A dataset is obtained by either, opening
the example dataset:

>>> import vaex
>>> t = vaex.example()

Or using :func:`open` or :func:`from_csv`, to open a file:

>>> t1 = vaex.open("somedata.hdf5")
>>> t2 = vaex.open("somedata.fits")
>>> t3 = vaex.from_csv("somedata.csv")

Or connecting to a remove server:

>>> tbig = vaex.open("http://bla.com/bigtable")

The main purpose of vaex is to provide statistics, such as mean, count, sum, standard deviation, per columns, possibly
with a selection, and on a regular grid.

To count the number of rows:

>>> t = vaex.example()
>>> t.count()
330000.0

Or the number of valid values, which for this dataset is the same:

>>> t.count("x")
330000.0

Count them on a regular grid:

>>> t.count("x", binby=["x", "y"], shape=(4,4))
array([[   902.,   5893.,   5780.,   1193.],
       [  4097.,  71445.,  75916.,   4560.],
       [  4743.,  71131.,  65560.,   4108.],
       [  1115.,   6578.,   4382.,    821.]])

Visualise it using matplotlib:

>>> t.plot("x", "y", show=True)
<matplotlib.image.AxesImage at 0x1165a5090>


"""# -*- coding: utf-8 -*-
from __future__ import print_function

try:
	from . import version
except:
	import sys
	print("version file not found, please run git/hooks/post-commit or git/hooks/post-checkout and/or install them as hooks (see git/README)", file=sys.stderr)
	raise

__version__ = version.versionstring
#__pre_release_name__ = version.pre_release
__version_tuple__ = version.versiontuple
__program_name__ = "vaex"
#__version_name__ = version.versiontring
#__release_name_ = version.versiontring[:]
#__clean_release__ = "%d.%d.%d" % (__version_tuple__)
__full_name__ = __program_name__ + "-" + __version__
#__clean_name__ =  __program_name__ + "-" + __clean_release__

__build_name__ = __full_name__ + "-" +version.osname


import glob
import vaex.dataset
from . import stat
#import vaex.file
#import vaex.export
from .delayed import delayed

import vaex.datasets
#import vaex.plot
#from vaex.dataset import Dataset
#del ServerRest, Dataset

def app(*args, **kwargs):
	"""Create a vaex app, the QApplication mainloop must be started.

	In ipython notebook/jupyter do the following:
	import vaex.ui.main # this causes the qt api level to be set properly
	import vaex as xs
	Next cell:
	%gui qt
	Next cell
	app = vx.app()

	From now on, you can run the app along with jupyter

	"""


	import vaex.ui.main
	return vaex.ui.main.VaexApp()
def open(path, *args, **kwargs):
	"""Open a dataset from file given by path

	:param str path: local or absolute path to file
	:param args: extra arguments for file readers that need it
	:param kwargs: extra keyword arguments
	:return: return dataset if file is supported, otherwise None
	:rtype: Dataset

	:Example:

	>>> import vaex as vx
	>>> vx.open('myfile.hdf5')
	<vaex.dataset.Hdf5MemoryMapped at 0x1136ee3d0>
	>>> vx.open('gadget_file.hdf5', 3) # this will read only particle type 3
	<vaex.dataset.Hdf5MemoryMappedGadget at 0x1136ef3d0>
	"""
	import vaex
	try:
		if path in aliases:
			path = aliases[path]
		if path.startswith("http://") or path.startswith("ws://"): # TODO: think about https and wss
			server, dataset = path.rsplit("/", 1)
			server = vaex.server(server, **kwargs)
			datasets = server.datasets(as_dict=True)
			if dataset not in datasets:
				raise KeyError("no such dataset '%s' at server, possible dataset names: %s" %  (dataset, " ".join(datasets.keys())))
			return datasets[dataset]
		if path.startswith("cluster"):
			import vaex.distributed
			return vaex.distributed.open(path, *args, **kwargs)
		else:
			import vaex.file
			return vaex.file.open(path, *args, **kwargs)
	except:
		logging.getLogger("vaex").error("error opening %r" % path)
		raise

def open_many(filenames):
	"""Open a list of filenames, and return a dataset with all datasets cocatenated

	:param list[str] filenames: list of filenames/paths
	:rtype: Dataset
	"""
	datasets = []
	for filename in filenames:
		filename = filename.strip()
		if filename and filename[0] != "#":
			datasets.append(open(filename))
	return vaex.dataset.DatasetConcatenated(datasets=datasets)

def from_samp(username=None, password=None):
	"""Connect to a SAMP Hub and wait for a single table load event, disconnect, download the table and return the dataset

	Useful if you want to send a single table from say TOPCAT to vaex in a python console or notebook
	"""
	print("Waiting for SAMP message...")
	import vaex.samp
	t = vaex.samp.single_table(username=username, password=password)
	return from_astropy_table(t.to_table())


def from_astropy_table(table):
	import vaex.file.other
	return vaex.file.other.DatasetAstropyTable(table=table)

def from_items(*items):
	"""Create an in memory dataset from numpy arrays, in contrast to from_arrays this keeps the order of columns intact

	:param: items: list of [(name, numpy array), ...]

	:Example:
	>>> x = np.arange(10)
	>>> y = x ** 2
	>>> dataset = vx.from_items(('x', x), ('y', y))


	"""
	dataset = vaex.dataset.DatasetArrays("array")
	for name, array in items:
		dataset.add_column(name, array)
	return dataset

def from_arrays(**arrays):
	"""Create an in memory dataset from numpy arrays


	:param: arrays: keyword arguments with arrays

	:Example:
	>>> x = np.arange(10)
	>>> y = x ** 2
	>>> dataset = vx.from_arrays(x=x, y=y)


	"""
	dataset = vaex.dataset.DatasetArrays("array")
	for name, array in arrays.items():
		dataset.add_column(name, array)
	return dataset

def from_scalars(**kwargs):
	"""Similar to from_arrays, but convenient for a dataset of length 1

	>>> ds = vx.from_scalars(x=1, y=2)
	"""
	import numpy as np
	return from_arrays( **{k:np.array([v]) for k, v in kwargs.items()})


def from_pandas(df, name="pandas", copy_index=True, index_name="index"):
	"""Create an in memory dataset from a pandas dataframe

	:param: pandas.DataFrame df: Pandas dataframe
	:param: name: unique for the dataset

	>>> import pandas as pd
	>>> df = pd.from_csv("test.csv")
	>>> ds = vx.from_pandas(df, name="test")
	"""
	import six
	dataset = vaex.dataset.DatasetArrays(name)
	def add(name, column):
		values = column.values
		if isinstance(values[0], six.string_types):
			values = values.astype("S")
		try:
			dataset.add_column(name, values)
		except Exception as e:
			print("could not convert column %s, error: %r, will try to convert it to string" % (name, e))
			try:
				values = values.astype("S")
				dataset.add_column(name, values)
			except Exception as e:
				print("Giving up column %s, error: %r" (name, e))
	for name in df.columns:
		add(name, df[name])
	if copy_index:
		add(index_name, df.index)
	return dataset

def from_ascii(path, seperator=None, names=True, skip_lines=0, skip_after=0, **kwargs):
	"""
	Create an in memory dataset from an ascii file (whitespace seperated by default).

	>>> ds = vx.from_ascii("table.asc")
	>>> ds = vx.from_ascii("table.csv", seperator=",", names=["x", "y", "z"])

	:param path: file path
	:param seperator: value seperator, by default whitespace, use "," for comma seperated values.
	:param names: If True, the first line is used for the column names, otherwise provide a list of strings with names
	:param skip_lines: skip lines at the start of the file
	:param skip_after: skip lines at the end of the file
	:param kwargs:
	:return:
	"""

	import vaex.ext.readcol as rc
	ds = vaex.dataset.DatasetArrays(path)
	if names not in [True, False]:
		namelist = names
		names = False
	else:
		namelist = None
	data = rc.readcol(path, fsep=seperator, asdict=namelist is None, names=names, skipline=skip_lines, skipafter=skip_after, **kwargs)
	if namelist:
		for name, array in zip(namelist, data.T):
			ds.add_column(name, array)
	else:
		for name, array in data.items():
			ds.add_column(name, array)
	return ds

def from_csv(filename_or_buffer, **kwargs):
	"""Shortcut to read a csv file using pandas and convert to a dataset directly"""
	import pandas as pd
	return from_pandas(pd.read_csv(filename_or_buffer, **kwargs))

def read_csv(filepath_or_buffer, **kwargs):
	'''Alias to from_csv'''
	return from_csv(filenames, **kwargs)

def _convert_name(filenames, shuffle=False):
	'''Convert a filename (or list of) to a filename with .hdf5 and optionally a -shuffle suffix'''
	if not isinstance(filenames, (list, tuple)):
		filenames = [filenames]
	base = filenames[0]
	if shuffle:
		base += '-shuffle'
	if len(filenames) > 1:
		return base + "_and_{}_more.hdf5".format(len(filenames))
	else:
		return base + ".hdf5"

def read_csv_and_convert(path, shuffle=False, copy_index=True, **kwargs):
	'''Convert a path (or glob pattern) to a single hdf5 file, will open the hdf5 file if exists

	Example:
		>>> vaex.read_csv_and_convert('test-*.csv', shuffle=True)  # this may take a while
		>>> vaex.read_csv_and_convert('test-*.csv', shuffle=True)  # 2nd time it is instant

	:param str path: path of file or glob pattern for multiple files
	:param bool shuffle: shuffle dataset when converting to hdf5
	:param bool copy_index: by default pandas will create an index (row number), set to false if you want to drop that
	:param kwargs: parameters passed to pandas' read_cvs

	'''
	from concurrent.futures import ProcessPoolExecutor
	import pandas as pd
	filenames = glob.glob(path)
	if len(filenames) > 1:
		filename_hdf5 = _convert_name(filenames, shuffle=shuffle)
		filename_hdf5_noshuffle = _convert_name(filenames, shuffle=False)
		if not os.path.exists(filename_hdf5):
			if not os.path.exists(filename_hdf5_noshuffle):
				#with ProcessPoolExecutor() as executor:
				#	executor.submit(read_csv_and_convert, filenames, shuffle=shuffle, **kwargs)
				for filename in filenames:
					read_csv_and_convert(filename, shuffle=shuffle, copy_index=copy_index, **kwargs)
				ds = open_many([_convert_name(k, shuffle=shuffle) for k in filenames])
			else:
				ds = open(filename_hdf5_noshuffle)
			ds.export_hdf5(filename_hdf5, shuffle=shuffle)
		return open(filename_hdf5)
	else:
		filename = filenames[0]
		filename_hdf5 = _convert_name(filename, shuffle=shuffle)
		filename_hdf5_noshuffle = _convert_name(filename, shuffle=False)
		if not os.path.exists(filename_hdf5):
			if not os.path.exists(filename_hdf5_noshuffle):
				df = pd.read_csv(filename, **kwargs)
				ds = from_pandas(df, copy_index=copy_index)
			else:
				ds = open(filename_hdf5_noshuffle)
			ds.export_hdf5(filename_hdf5, shuffle=shuffle)
		return open(filename_hdf5)




import vaex.settings
aliases = vaex.settings.main.auto_store_dict("aliases")

# py2/p3 compatibility
try:
	from urllib.parse import urlparse
except ImportError:
	from urlparse import urlparse

def server(url, **kwargs):
	"""Connect to hostname supporting the vaex web api

	:param str hostname: hostname or ip address of server
	:return vaex.dataset.ServerRest: returns a server object, note that it does not connect to the server yet, so this will always succeed
	:rtype: ServerRest
	"""
	from vaex.remote import ServerRest
	url = urlparse(url)
	if url.scheme == "ws":
		websocket = True
	else:
		websocket = False
	assert url.scheme in ["ws", "http"]
	port = url.port
	base_path = url.path
	hostname = url.hostname
	return vaex.remote.ServerRest(hostname, base_path=base_path, port=port, websocket=websocket, **kwargs)

def example(download=True):
	"""Returns an example dataset which comes with vaex for testing/learning purposes

	:rtype: vaex.dataset.Dataset
	"""
	from . import utils
	path = utils.get_data_file("helmi-dezeeuw-2000-10p.hdf5")
	if path is None and download:
		return vaex.datasets.helmi_de_zeeuw_10percent.fetch()
	return open(path) if path else None


def zeldovich(dim=2, N=256, n=-2.5, t=None, scale=1, seed=None):
	"""Creates a zeldovich dataset
	"""
	import vaex.file
	return vaex.file.other.Zeldovich(dim=dim, N=N, n=n, t=t, scale=scale)

def set_log_level_debug():
	"""set log level to debug"""
	import logging
	logging.getLogger("vaex").setLevel(logging.DEBUG)

def set_log_level_info():
	"""set log level to info"""
	import logging
	logging.getLogger("vaex").setLevel(logging.INFO)

def set_log_level_warning():
	"""set log level to warning"""
	import logging
	logging.getLogger("vaex").setLevel(logging.WARNING)

def set_log_level_exception():
	"""set log level to exception"""
	import logging
	logging.getLogger("vaex").setLevel(logging.FATAL)

def set_log_level_off():
	"""Disabled logging"""
	import logging
	logging.disable(logging.CRITICAL)


import logging
format = "%(levelname)s:%(threadName)s:%(name)s:%(message)s"
logging.basicConfig(level=logging.INFO, format=format)
#logging.basicConfig(level=logging.DEBUG)
set_log_level_warning()
import os
import_script = os.path.expanduser("~/.vaex/vaex_import.py")
if os.path.exists(import_script):
	try:
		with open(import_script) as f:
			code = compile(f.read(), import_script, 'exec')
			exec(code)
	except:
		import traceback
		traceback.print_tb()

import logging
logger = logging.getLogger('vaex')

import pkg_resources
for entry in pkg_resources.iter_entry_points(group='vaex.namespace'):
	logger.debug('adding vaex namespace: ' + entry.name)
	try:
	    add_namespace = entry.load()
	    add_namespace()
	except Exception:
		logger.exception('issue loading ' + entry.name)

for entry in pkg_resources.iter_entry_points(group='vaex.plugin'):
	logger.debug('adding vaex plugin: ' + entry.name)
	try:
	    add_namespace = entry.load()
	    add_namespace()
	except Exception:
		logger.exception('issue loading ' + entry.name)



import vaex.utils
  
registry = {}

def register(cls):
    registry[fullname(cls)] = cls
    return cls

def fullname(cls):
    return cls.__module__ + '.' + cls.__name__

def to_dict(obj):
    return dict(cls=fullname(obj.__class__), state=obj.state_get())

def from_dict(d):
    cls_name = d['cls']
    if cls_name not in registry:
        raise ValueError('unknown class: ' + cls_name)
    else:
        obj = registry[cls_name].state_from(d['state'])
        #obj.state_set(d['state'])
        return obj

def can_serialize(object):
    return fullname(object.__class__) in registry